from __future__ import annotations

import abc
import requests
import json
import logging
import jwt
import importlib.metadata
from datetime import datetime
from concurrent.futures import Future
from typing import Union, Any, Optional, List, Dict
from pydantic import ValidationError, AnyUrl, PydanticUserError

from airadar.utils import get_git_info
from airadar.utils.config import airadar_configs, ConfigKeys
from airadar.utils.auth import get_ccf_access_token, get_private_key_access_token
from airadar.typing import RadarResponse
from airadar.tracking.model import (
    PipelineRunSchema,
    PipelineRunCreateRequestSchema,
    DatasetVersionCreateRequestSchema,
    DatasetSchema,
    ModelVersionCreateRequestSchema,
    ModelSchema,
    Dependency,
    File,
    PipelineSchema,
    PipelineCreateRequestSchema,
    PipelineRunState,
    PipelineRunUpdateRequestSchema,
    DeploymentCreateRequestSchema,
    DeploymentWithApplicationSchema,
    ExternalRunUrlItem,
    ApplicationUrlItem,
    BuildUrlItem,
    ExternalRunIdItem,
    DescriptionItem,
    RadarUri,
    VersionItem3,
    VersionItem,
    ExternalUrlItem,
)

logger = logging.getLogger(__name__)

try:
    RADAR_SDK_VERSION = importlib.metadata.version("airadar")
except importlib.metadata.PackageNotFoundError:
    RADAR_SDK_VERSION = "unknown"


class MissingAuthException(Exception):
    """Raised when the environment variables for obtaining access tokens are not set properly"""


class InvalidAuthTokenException(Exception):
    """Raised when we fail to obtain an access token from the auth provider"""


class AccessToken:
    def __init__(self, access_token: str):
        try:
            decoded_token = jwt.decode(
                access_token, options={"verify_signature": False}
            )
            self.expiration_timestamp = decoded_token.get("exp")
        except jwt.DecodeError:
            logger.critical("Failed to decode JWT access token")
            raise InvalidAuthTokenException("Access token is not valid")

        self.access_token = access_token

    def is_expired(self) -> bool:
        if not self.access_token or not self.expiration_timestamp:
            raise InvalidAuthTokenException("Access token is not valid")

        expiration_datetime = datetime.utcfromtimestamp(self.expiration_timestamp)
        return datetime.utcnow() > expiration_datetime


class HTTPClient(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        """
        Base class for api clients
        """
        self.base_url = str(airadar_configs.get_value(ConfigKeys.AIRADAR_API_SERVER))
        self.api_version_str = (
            airadar_configs.get_value(ConfigKeys.AIRADAR_API_VERSION) or "v1"
        )
        self.base_url = self.base_url.rstrip("/")

        self.headers = {
            "Content-Type": "application/json",
            "User-Agent": f"radar-sdk-python/{RADAR_SDK_VERSION}",
        }

    @abc.abstractmethod
    def post(self, end_point: str, json_data: str) -> Union[Future[Any], RadarResponse]:
        """
        Abstract method for a HTTP POST call. Derived classes should implement this method.

        Arguments:
            end_point: End-point at the base_url
            json_data: JSON payload

        Returns:
            Either an [APIResponse][airadar.api.api_client.APIResponse] or a Future in case of async request.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def patch(
        self, end_point: str, json_data: str
    ) -> Union[Future[Any], RadarResponse]:
        """
        Abstract method for a HTTP PATCH call. Derived classes should implement this method.

        Arguments:
            end_point: End-point at the base_url
            json_data: JSON payload

        Returns:
            Either an [APIResponse][airadar.api.api_client.APIResponse] or a Future in case of async request.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, end_point: str) -> Union[Future[Any], RadarResponse]:
        """
        Abstract method for a HTTP GET call. Derived classes should implement this method.

        Arguments:
            end_point: End-point at the base_url

        Returns:
            Either an [APIResponse][airadar.api.api_client.APIResponse] or a Future in case of async request.
        """
        raise NotImplementedError


class SyncHTTPClient(HTTPClient):
    def __init__(self) -> None:
        super().__init__()
        self.jwt_access_token: AccessToken = SyncHTTPClient.fetch_access_token()

    @staticmethod
    def fetch_access_token() -> AccessToken:
        auth_client_url = airadar_configs.get_value(ConfigKeys.AIRADAR_AUTH_TENANT_URL)
        auth_client_id = airadar_configs.get_value(ConfigKeys.AIRADAR_AUTH_CLIENT_ID)
        auth_client_secret = airadar_configs.get_value(
            ConfigKeys.AIRADAR_AUTH_CLIENT_SECRET
        )
        auth_private_key = airadar_configs.get_value(
            ConfigKeys.AIRADAR_AUTH_CLIENT_PRIVATE_KEY
        )
        auth_key_id = airadar_configs.get_value(ConfigKeys.AIRADAR_AUTH_CLIENT_KID)
        auth_client_scope = airadar_configs.get_value(
            ConfigKeys.AIRADAR_AUTH_CLIENT_SCOPE
        )

        if (
            auth_client_url is None
            or auth_client_id is None
            or auth_client_scope is None
        ):
            raise MissingAuthException("Auth configurations are missing")

        # Prefer private key auth, otherwise fallback to client credential flow
        if auth_private_key is not None and auth_key_id is not None:
            logger.debug("Using private key auth strategy")
            access_token = get_private_key_access_token(
                str(auth_client_url),
                str(auth_client_id),
                str(auth_private_key),
                str(auth_key_id),
                str(auth_client_scope),
            )
        elif auth_client_secret is not None:
            logger.debug("Using client credential flow auth strategy")
            access_token = get_ccf_access_token(
                str(auth_client_url),
                str(auth_client_id),
                str(auth_client_secret),
                str(auth_client_scope),
            )
        else:
            raise MissingAuthException("Auth configurations are missing")

        if access_token is None:
            logger.critical(
                f"""
                Failed to obtain an access token. Check the logs for more information. 
                Make sure env variables are properly set.
                """
            )
            raise InvalidAuthTokenException("Failed to obtain an access token")

        logger.debug("Obtained a valid access token from auth provider")
        return AccessToken(access_token)

    def _handle_http_method(
        self, method: str, end_point: str, json_data: Optional[str] = None
    ) -> RadarResponse:
        try:
            logger.debug("Checking for a valid access token")
            if not self.jwt_access_token:
                self.jwt_access_token = SyncHTTPClient.fetch_access_token()
            elif self.jwt_access_token.is_expired():
                self.jwt_access_token = SyncHTTPClient.fetch_access_token()
        except (InvalidAuthTokenException, MissingAuthException) as e:
            error_msg = f"Failed to make {method} call due to invalid access token: {e}"
            logger.warning(error_msg)
            return RadarResponse(error_msg=error_msg)

        end_point = end_point.lstrip("/")
        url = f"{self.base_url}/{self.api_version_str}/{end_point}"
        logger.debug("Initiating %s request to endpoint %s", method, url)

        # urllib allows opening both http:// and file:// URLs. To avoid the security risk of
        # accidentally opening the latter, we need to validate the URL is an HTTP endpoint
        if not url.lower().startswith("http"):
            error_msg = f"Invalid URL {url} passed to post_json, skipping request"
            logger.warning(error_msg)
            return RadarResponse(error_msg=error_msg)

        func = None
        try:
            func = getattr(requests, method)
        except AttributeError as ae:
            error_msg = f"HTTP requested method is not part of requests library. {ae}"
            logger.exception(error_msg)
            return RadarResponse(error_msg=error_msg)

        headers = {
            **self.headers,
            "Authorization": f"Bearer {self.jwt_access_token.access_token}",
        }
        try:
            resp = None
            if json_data:
                resp = func(
                    url=url,
                    json=json.loads(json_data),
                    headers=headers,
                    timeout=10,
                )
            else:
                resp = func(url=url, headers=self.headers, timeout=10)

            resp.raise_for_status()
            resp_json = resp.json()
            entity_uuid = resp_json.get("uuid")
            if entity_uuid is None:
                error_msg = f"Could not get entity UUID from response body: {resp_json}"
                logger.error(error_msg)
                return RadarResponse(
                    http_status_code=resp.status_code, error_msg=error_msg
                )

            return RadarResponse(
                http_status_code=resp.status_code,
                entity_uuid=entity_uuid,
                json_response=dict(resp_json),
            )
        except json.decoder.JSONDecodeError as e:
            error_msg = f"Error parsing json payload for request to {url}: {str(e)}"
            logger.error(error_msg)
            return RadarResponse(error_msg=error_msg)
        except requests.exceptions.HTTPError as e:
            if e.response is None:
                return RadarResponse(error_msg="Unknown HTTP Error occurred")
            error_msg = str(
                f"Error with request: {e.response.status_code}\n{e.response.text}"
            )
            logger.error(error_msg)
            return RadarResponse(
                http_status_code=e.response.status_code, error_msg=error_msg
            )
        except requests.exceptions.JSONDecodeError as e:
            error_msg = f"Response was not valid JSON for request to {url}: {str(e)}"
            logger.error(error_msg)
            return RadarResponse(http_status_code=500, error_msg=error_msg)
        except requests.exceptions.Timeout as e:
            error_msg = f"Request timed out for request to {url}"
            logger.error(error_msg)
            return RadarResponse(http_status_code=408, error_msg=error_msg)
        except Exception as e:
            error_msg = f"Unexpected error during request to {url}: {str(e)}"
            logger.error(error_msg)
            status_code = None
            if (
                hasattr(e, "response")
                and hasattr(e.response, "status_code")
                and isinstance(e.response.status_code, int)
            ):
                status_code = e.response.status_code
            return RadarResponse(http_status_code=status_code, error_msg=error_msg)

    def post(self, end_point: str, json_data: str) -> RadarResponse:
        return self._handle_http_method("post", end_point, json_data)

    def patch(self, end_point: str, json_data: str) -> RadarResponse:
        return self._handle_http_method("patch", end_point, json_data)

    def get(self, end_point: str) -> RadarResponse:
        return self._handle_http_method("get", end_point)


class RadarAPIClient:
    def __init__(self) -> None:
        self.http_client = self._setup_http_client()

    def _setup_http_client(self) -> Union[SyncHTTPClient, None]:
        try:
            return SyncHTTPClient()
        except (InvalidAuthTokenException, MissingAuthException):
            pass
        except Exception as e:
            logger.critical(
                "Unexpected error when fetching Radar access token: %s", str(e)
            )

        return None

    def is_authorized(self) -> bool:
        return self.http_client is not None

    def _validate_api_response(
        self,
        response: RadarResponse,
        model_class: Any,
    ) -> RadarResponse:
        try:
            model_class.model_validate(response.json_response)
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=response.http_status_code,
                error_msg=error_msg,
                json_response=response.json_response,
            )

        return response

    def post_pipeline(
        self, pipeline_name: str, external_url: Optional[ExternalUrlItem] = None
    ) -> RadarResponse:
        if self.http_client is None:
            return RadarResponse(error_msg="Client data collection disabled")

        try:
            pipeline_create_obj = PipelineCreateRequestSchema(
                name=pipeline_name, external_url=external_url
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Failed to pipeline create request object."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        pipeline_create_res = self.http_client.post(
            "/pipelines",
            pipeline_create_obj.model_dump_json(),
        )

        if not pipeline_create_res.is_success():
            logger.critical(
                "AI Radar pipeline could not be created, no data will be collected by AI Radar for this run"
            )
            return pipeline_create_res

        try:
            pipeline = PipelineSchema.model_validate(pipeline_create_res.json_response)
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for pipeline create request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            self.disable_data_collection = True
            return RadarResponse(
                http_status_code=pipeline_create_res.http_status_code,
                error_msg=error_msg,
                json_response=pipeline_create_res.json_response,
            )

        # Set the pipeline run id here to avoid duplicate calls when airadar.log_artifact() is called
        # following a call to airadar.active_run()
        self.pipeline_id = pipeline.uuid
        self.pipeline = pipeline
        return pipeline_create_res

    def post_pipeline_run(
        self,
        pipeline_id: str,
        external_run_id: Optional[str] = None,
        run_at: Optional[datetime] = None,
        run_uri: Optional[str] = None,
        dependency_report: List[Dependency] = [],
    ) -> RadarResponse:
        if self.http_client is None:
            return RadarResponse(error_msg="Client data collection disabled")

        git_info = get_git_info()

        try:
            # TODO: Populate platform with a meaningful name
            # See https://github.com/protectai/radar-sdk/issues/72
            create_run_obj = PipelineRunCreateRequestSchema(
                external_run_id=(
                    ExternalRunIdItem(root=external_run_id) if external_run_id else None
                ),
                external_run_url=(
                    ExternalRunUrlItem(root=AnyUrl(run_uri)) if run_uri else None
                ),
                dependencies=dependency_report,
                run_at=run_at or datetime.now(),
                git_info=git_info,
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Error when trying to create request based on given inputs."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        run_create_res = self.http_client.post(
            f"/pipelines/{pipeline_id}/runs",
            create_run_obj.model_dump_json(),
        )

        if not run_create_res.is_success():
            logger.critical(
                "AI Radar pipeline run ID could not be retrieved, no data will be collected by AI Radar for this run"
            )
            return run_create_res

        try:
            pipeline_run = PipelineRunSchema.model_validate(
                run_create_res.json_response
            )
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for pipeline run create request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            self.disable_data_collection = True
            return RadarResponse(
                http_status_code=run_create_res.http_status_code,
                error_msg=error_msg,
                json_response=run_create_res.json_response,
            )

        # Set the pipeline run id here to avoid duplicate calls when airadar.log_artifact() is called
        # following a call to airadar.active_run()
        self.pipeline_run_id = pipeline_run.uuid
        self.pipeline_run = pipeline_run
        return run_create_res

    def patch_pipeline_run(
        self,
        status: PipelineRunState,
        pipeline_run_id: str,
        completed_at: Optional[datetime] = None,
    ) -> RadarResponse:
        if self.http_client is None:
            return RadarResponse(error_msg="Client data collection disabled")

        try:
            patch_run_obj = PipelineRunUpdateRequestSchema(
                status=status, completed_at=completed_at
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Error when trying to create request based on given inputs."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        run_patch_res = self.http_client.patch(
            f"/pipelines/run/{pipeline_run_id}",
            patch_run_obj.model_dump_json(),
        )

        if not run_patch_res.is_success():
            logger.critical(
                f"Failed to mark pipeline run {pipeline_run_id} complete in AI Radar"
            )
            return run_patch_res

        try:
            patch_response = PipelineRunSchema.model_validate(
                run_patch_res.json_response
            )
            logger.info(
                f"Pipeline run with uuid={patch_response.uuid} updated successfully"
            )
            return run_patch_res
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for pipeline run patch request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=run_patch_res.http_status_code,
                error_msg=error_msg,
                json_response=run_patch_res.json_response,
            )

    def post_model(
        self,
        name: str,
        uri: str,
        files: List[File],
        version: Optional[str] = None,
        pipeline_run_id: Optional[str] = None,
        description: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> RadarResponse:
        if not self.http_client:
            return RadarResponse(error_msg="Client data collection disabled")

        if pipeline_run_id:
            model_version_endpoint = f"/pipelines/run/{pipeline_run_id}/model-versions"
        else:
            model_version_endpoint = "/models"

        try:
            request = ModelVersionCreateRequestSchema(
                name=name,
                external_uri=uri,
                version=VersionItem3(root=version) if version else None,
                description=DescriptionItem(root=description) if description else None,
                files=files,
                properties=properties,
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Error when trying to create request based on given inputs."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        response = self.http_client.post(
            model_version_endpoint, request.model_dump_json(by_alias=True)
        )

        if not response.is_success():
            logger.critical(f"Failed to create model version for {name}")
            return response

        try:
            model_response = ModelSchema.model_validate(response.json_response)
            logger.info(f"Model version created with uuid={model_response.uuid}")
            return response
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for model version create request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=response.http_status_code,
                error_msg=error_msg,
                json_response=response.json_response,
            )

    def post_input_model(self, radar_uri: str, pipeline_run_id: str) -> RadarResponse:
        if not self.http_client:
            return RadarResponse(error_msg="Client data collection disabled")

        associate_model_endpoint = f"/pipelines/run/{pipeline_run_id}/associate-model"

        response = self.http_client.post(
            associate_model_endpoint, json.dumps({"radar_uri": radar_uri})
        )

        if not response.is_success():
            logger.critical(
                f"Failed to associate input model version at {radar_uri} with pipeline run {pipeline_run_id}"
            )
            return response

        try:
            model_response = ModelSchema.model_validate(response.json_response)
            if model_response.latest_version:
                logger.info(
                    f"Model version uuid={model_response.latest_version.uuid} was successfully associated with pipeline run {pipeline_run_id}"
                )
            return response
        except ValidationError as e:
            error_msg = f"Got an invalid response from the API server when trying to associate input model {radar_uri}"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=response.http_status_code,
                error_msg=error_msg,
                json_response=response.json_response,
            )

    def post_dataset(
        self,
        name: str,
        uri: str,
        files: List[File],
        pipeline_run_id: str,
        version: Optional[str] = None,
        description: Optional[str] = None,
    ) -> RadarResponse:
        if not self.http_client:
            return RadarResponse(error_msg="Client data collection disabled")

        dataset_version_endpoint = f"/pipelines/run/{pipeline_run_id}/dataset-versions"

        try:
            request = DatasetVersionCreateRequestSchema(
                name=name,
                external_uri=uri,
                version=VersionItem(root=version) if version else None,
                description=DescriptionItem(root=description) if description else None,
                files=files,
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Error when trying to create request based on given inputs."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        response = self.http_client.post(
            dataset_version_endpoint, request.model_dump_json()
        )

        if not response:
            logger.critical(f"Failed to create dataset version for {name}")
            return response

        try:
            dataset_response = DatasetSchema.model_validate(response.json_response)
            logger.info(f"Dataset version created with uuid={dataset_response.uuid}")
            return response
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for dataset version create request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=response.http_status_code,
                error_msg=error_msg,
                json_response=response.json_response,
            )

    def post_deployment(
        self,
        application_id: str,
        application_version: str,
        environment: str,
        application_url: Optional[str] = None,
        build_url: Optional[str] = None,
        radar_uris: Optional[List[str]] = None,
        build_time: datetime = datetime.now(),
    ) -> RadarResponse:
        if not self.http_client:
            return RadarResponse(error_msg="Client data collection disabled")

        try:
            deployment = DeploymentCreateRequestSchema(
                application_version=application_version,
                environment=environment,
                application_url=(
                    ApplicationUrlItem(root=AnyUrl(application_url))
                    if application_url
                    else None
                ),
                build_url=BuildUrlItem(root=AnyUrl(build_url)) if build_url else None,
                build_time=build_time,
                radar_uris=(
                    [RadarUri(root=radar_uri) for radar_uri in radar_uris]
                    if radar_uris
                    else []
                ),
            )
        except (ValidationError, PydanticUserError) as e:
            error_msg = "Error when trying to create request based on given inputs."
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(error_msg=error_msg)

        response = self.http_client.post(
            f"/applications/{application_id}/deployments",
            deployment.model_dump_json(),
        )
        if not response.is_success():
            logger.critical(
                f"Failed to create a deployment for application id {application_id}"
            )
            return response

        try:
            deployment_response = DeploymentWithApplicationSchema.model_validate(
                response.json_response
            )
            logger.info(f"Deployment created with uuid={deployment_response.uuid}")
            return response
        except ValidationError as e:
            error_msg = "Got an invalid response from the API server for deployment create request"
            logger.critical(
                error_msg,
                e,
                exc_info=True,
            )
            return RadarResponse(
                http_status_code=response.http_status_code,
                error_msg=error_msg,
                json_response=response.json_response,
            )



from typing import Union, List, Dict, Optional, Tuple
from types import FrameType
import os
import sys
from git import InvalidGitRepositoryError
from git.repo import Repo
from pathlib import Path
from urllib.parse import urlparse
from glob import glob

from airadar.tracking.model import (
    GitInfo,
    ShaItem,
    RepoUrlItem,
    RepoNameItem,
    RepoTypeItem,
    File,
    Dependency,
    License,
    LicenseWrapper,
)
from airadar.tools.cloud_storage import CloudStorage
from airadar.extractors.dependency import (
    DependencyExtractor,
    PackageDependencyType,
)
import logging

logger = logging.getLogger(__name__)


def get_git_info() -> GitInfo:
    git_info = GitInfo()
    try:
        repo = Repo(search_parent_directories=True)
        sha = str(repo.head.object.hexsha)

        git_info.sha = ShaItem(root=sha)
        git_info.repo_url = RepoUrlItem(root=repo.remotes.origin.url)
        git_info.repo_name = RepoNameItem(
            root=repo.remotes.origin.url.split(".git")[0].split("/")[-1]
        )
        git_info.repo_type = RepoTypeItem(root="git")
        git_info.dirty = repo.is_dirty()
    except InvalidGitRepositoryError:
        logger.info("Not a valid git repo, no git metadata will be recorded")
        git_info.repo_type = RepoTypeItem(root="Unknown")
    except Exception as e:
        logger.warning("Failed to capture git context: %s", str(e))
        git_info.repo_type = RepoTypeItem(root="Unknown")

    return git_info


def get_calling_function_name() -> Union[str, None]:
    radar_logging_functions = [
        "log_artifact",
        "log_dataset",
        "log_model",
    ]
    try:
        frame: Optional[FrameType] = sys._getframe(0) or None
        in_radar_call_stack = False
        while frame:
            # We are within radar SDK function stack and can terminate the search as soon we reached
            # out of it.
            if (
                in_radar_call_stack
                and frame.f_code.co_name not in radar_logging_functions
            ):
                break

            # we are in user functions stack and will keep looking until a radar specific function is
            # found.
            if (
                not in_radar_call_stack
                and frame.f_code.co_name in radar_logging_functions
            ):
                in_radar_call_stack = True

            frame = frame.f_back

        if frame:
            # If code is executed in a Python file outside of a function, just use the filename
            if frame.f_code.co_name == "<module>":
                return os.path.basename(frame.f_code.co_filename)
            return frame.f_code.co_name
    except Exception as e:
        logger.warning("Failed to get calling function name: %s", str(e))

    return None


def name_artifact(artifact_path: Path) -> str:
    # Handle glob cases first - find nearest parent directory that contains no wildcards
    path_str = str(artifact_path)
    if "*" in path_str:
        parts = path_str.split("/")
        for part in reversed(parts):
            if "*" not in part and len(part) > 0:
                return part

    # Returns the parent directory or filename of Path
    return artifact_path.name


def update_file_metadata(files: List[File]) -> None:
    for file in files:
        if file.uri is None:
            continue

        storage_obj = CloudStorage.get_storage_obj_for_uri(file.uri)
        if not storage_obj:
            continue

        file_metadata = storage_obj.get_file_metadata(file.uri)
        if file_metadata is None:
            continue

        # Do not overwrite explicitly set version
        if file.version is None:
            file.version = file_metadata.version
            file.version_strategy = file_metadata.version_strategy

        file.resource_name = file_metadata.resource_name
        file.storage_type = file_metadata.storage_type
        file.size = file_metadata.size


def get_files_and_name_for_uri(
    uri: str,
) -> Tuple[List[File], Optional[str]]:
    artifact_list: List[str] = []
    inferred_name = None

    parsed_uri = urlparse(uri)

    storage_obj = CloudStorage.get_storage_obj_for_uri(uri)
    if storage_obj:
        inferred_name = name_artifact(Path(parsed_uri.path))
        file_list = storage_obj.get_files_for_uri(uri)
        if file_list is None:
            error_msg = f"Could not fetch file listing in cloud storage path: {uri}. Check radar.log for more details."
            logger.warning(error_msg)
        else:
            artifact_list.extend(file_list)
    else:
        if parsed_uri.scheme == "file" or not parsed_uri.scheme:
            root_dir = Path(parsed_uri.path)
            inferred_name = name_artifact(root_dir)

            if root_dir.is_file():
                artifact_list.append(str(root_dir))

            # If the user has provided a directory path, treat all contents of dir as matching
            elif root_dir.exists() and root_dir.is_dir():
                glob_list = root_dir.glob("**/*")
                matching_list = [str(artifact) for artifact in glob_list]
                # Filter for only files, not directories
                for match in matching_list:
                    if Path(match).is_file():
                        artifact_list.append(match)

            # Get list of matching files from user's input glob
            else:
                matching_list = glob(parsed_uri.path, recursive=True)
                # Filter for only files, not directories
                for match in matching_list:
                    if Path(match).is_file():
                        artifact_list.append(match)
        elif parsed_uri.scheme in ["runs", "models"]:
            # For now we just assume mlflow URIs, we can add additional check to detect model registry type with additional integs.
            # TODO: Check MLFLOW_TRACKING_URI is set

            from airadar.tools.model_registry import get_mlflow_model_info
            from airadar.tools.model_registry import get_mlflow_model_files

            model_version = get_mlflow_model_info(uri)
            if model_version:
                inferred_name = model_version.name
            else:
                inferred_name = "mlflow_model"

            if model_version:
                file_paths = get_mlflow_model_files(model_version.run_id)
                if file_paths:
                    artifact_list.extend(file_paths)
        else:
            logger.info(
                f"Radar currently doesn't know how to handle {uri}. We might only have limited metadata for this URI."
            )

    files: List[File] = []
    for path in artifact_list:
        files.append(File(uri=path))

    update_file_metadata(files)

    return files, inferred_name


# Create list of licenses in correct format and truncate names to 128 chars if longer.
def generate_license_list(licenses: List[Dict[str, str]]) -> List[LicenseWrapper]:
    return [
        LicenseWrapper(license=License(name=value[:128]))
        for license in licenses
        for key, value in license.items()
        if key == "name"
    ]


def get_dependencies_from_environment() -> List[Dependency]:
    dependency_list = []

    extractor = DependencyExtractor(PackageDependencyType.ENVIRONMENT)
    dependency_report = extractor.get_report()

    # Create dependency list with truncated values so users don't receive API pydcantic validation errors.
    dependency_components = dependency_report.get("components", [])
    for dependency_component in dependency_components:
        if dependency_component.get(
            "type", ""
        ) == "library" and dependency_component.get("name", ""):
            dependency_list.append(
                Dependency(
                    name=dependency_component.get("name", "")[:128],
                    version=dependency_component.get("version", "")[:256],
                    author=dependency_component.get("author", "")[:256],
                    purl=dependency_component.get("purl", "")[:256],
                    licenses=generate_license_list(
                        dependency_component.get("licenses", [])
                    ),
                )
            )

    return dependency_list



# mypy: disable-error-code="assignment,attr-defined"

from typing import TypeVar, Optional, Dict, Any, Union

# Annotated is only in typing module as of Python 3.9, earlier
# versions require importing from typing_extensions. Depending on the version of
# Python used, one line or the other will cause a type error, so we ignore both
try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated


class RadarResponse:
    """
    Represents a response from interacting with the Radar SDK.

    Arguments:
        http_status_code: Status code returned from the API, or None if no request to the API was made
        entity_uuid: UUID representing the entity that was created or modified in the API call if the request was
            successful, or None if the call failed
        json_response: JSON response data from the API specific to the created or updated entity if the request
            was made, or None if the call failed
        error_msg: Detailed error message if there was an error in the Radar SDK or API, or None if the call succeeded
    """

    def __init__(
        self,
        http_status_code: Optional[int] = None,
        entity_uuid: Optional[str] = None,
        json_response: Optional[Dict[str, Any]] = None,
        error_msg: Optional[str] = None,
    ):
        self._status_code = http_status_code
        self._entity_uuid = entity_uuid
        self._json_response = json_response
        self._error_msg = error_msg

    @property
    def http_status_code(self) -> Optional[int]:
        return self._status_code

    @property
    def entity_uuid(self) -> Optional[str]:
        return self._entity_uuid

    @property
    def json_response(self) -> Optional[Dict[str, Any]]:
        return self._json_response

    @property
    def error_msg(self) -> Optional[str]:
        return self._error_msg

    @error_msg.setter
    def error_msg(self, msg: str) -> None:
        self._error_msg = msg

    def is_success(self) -> bool:
        """
        Helper method to tell if Radar request succeeded or not.

        Returns:
            True if the call succeeded, False otherwise
        """
        return self.entity_uuid is not None and self.error_msg is None


class BaseType:
    pass


class ArtifactType(BaseType):
    """
    Represents a generic machine learning artifact.

    Use one of the specific type of artifact radar.Model or radar.Dataset

    Arguments:
        type: Type of the artifact
        uri: The artifact's location on disk or cloud storage.
        metadata: Arbitrary key-value pairs describing the artifact.

    """

    def __init__(
        self,
        type: Optional[str] = None,
        uri: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.type = type
        self.uri = uri
        self.metadata = metadata


class DatasetType(ArtifactType):
    """
    Represent a machine learning dataset.

    Examples:
        ```
        @radar.pipeline(name="feature_generation")
        def feature_generation(dataset: radar.Dataset[pd.Dataframe]):
            ...
        ```
    """


class ModelType(ArtifactType):
    """
    Represent a machine learning model.
    """


class InputAnnotation:
    """A type to mark input artifact"""


class OutputAnnotation:
    """A type to mark output artifact"""


T = TypeVar("T")
TT = TypeVar("TT")

Artifact = Annotated[T, ArtifactType]
"""
A type to mark a variable as artifact

Arguments:
    T: Any type
"""


Dataset = Annotated[T, DatasetType]
"""
A type to mark a variable as dataset

Arguments:
    T: Any type

Examples:

    ```
    import pandas as pd

    @radar.pipeline(name="training")
    def training(train_data: radar.Dataset[pd.DataFrame])
        ...
    ```
"""

Model = Annotated[T, ModelType]
"""
A type to mark a variable as model

Arguments:
    T: Any type

Examples:

    ```
    import pandas as pd

    @radar.pipeline(name="training")
    def training(train_data: radar.Dataset[pd.DataFrame]) -> radar.Model
        ...
    ```
"""

Input = Annotated[TT, InputAnnotation]
"""
A type to indicate that the marked artifact is input.

Arguments:
    TT: ArtifactType, DatasetType, or ModelType
"""


Output = Annotated[TT, OutputAnnotation]
"""
A type to indicate that the marked artifact is input.

Arguments:
    TT: ArtifactType, DatasetType, or ModelType
"""

# Helper type for representing any of the 3 artifact types
AnyArtifact = Union[Dataset[Any], Model[Any]]



# generated by datamodel-codegen:
#   filename:  http://localhost:3001/openapi.json
#   timestamp: 2024-09-11T17:38:42+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Dict, List, Optional, Union

from airadar.utils.base_model import RadarBaseModel
from pydantic import AnyUrl, ConfigDict, Field, RootModel


class ExternalUrlItem(RootModel[AnyUrl]):
    root: AnyUrl


class DescriptionItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class ApplicationCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(max_length=128, title="Name")]
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class ApplicationStatsSchema(RadarBaseModel):
    total_applications: Annotated[int, Field(title="Total Applications")]
    applications_with_models_failing_model_scan: Annotated[
        str, Field(max_length=128, title="Applications With Models Failing Model Scan")
    ]
    applications_with_models_missing_model_scan: Annotated[
        str, Field(max_length=128, title="Applications With Models Missing Model Scan")
    ]


class NameItem(RootModel[str]):
    root: Annotated[str, Field(max_length=128)]


class ApplicationUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[Optional[NameItem], Field(title="Name")] = None
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class ArtifactDirection(Enum):
    input = "INPUT"
    output = "OUTPUT"


class CVESeverity(RadarBaseModel):
    rating: Annotated[str, Field(max_length=256, title="Rating")]
    score: Annotated[float, Field(title="Score")]
    vector: Annotated[str, Field(max_length=256, title="Vector")]


class DatasetBaseSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class DatasetStatsSchema(RadarBaseModel):
    total_datasets: Annotated[int, Field(title="Total Datasets")]
    total_related_models: Annotated[int, Field(title="Total Related Models")]
    total_related_model_versions: Annotated[
        int, Field(title="Total Related Model Versions")
    ]


class DatasetUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[Optional[NameItem], Field(title="Name")] = None
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class VersionItem(RootModel[str]):
    root: Annotated[str, Field(max_length=64)]


class DatasetVersionWithDatasetSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    dataset: DatasetBaseSchema
    version: Annotated[str, Field(max_length=64, title="Version")]
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    file_count: Annotated[int, Field(title="File Count")]


class VersionItem1(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class AuthorItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class PurlItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class ApplicationUrlItem(RootModel[AnyUrl]):
    root: AnyUrl


class BuildUrlItem(RootModel[AnyUrl]):
    root: AnyUrl


class RadarUri(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class DeploymentCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    application_version: Annotated[
        str, Field(max_length=256, title="Application Version")
    ]
    environment: Annotated[str, Field(max_length=256, title="Environment")]
    application_url: Annotated[
        Optional[ApplicationUrlItem], Field(title="Application Url")
    ] = None
    build_url: Annotated[Optional[BuildUrlItem], Field(title="Build Url")] = None
    build_time: Annotated[datetime, Field(title="Build Time")]
    radar_uris: Annotated[List[RadarUri], Field(title="Radar Uris")]


class ApplicationVersionItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class EntityType(Enum):
    pipeline = "PIPELINE"
    application = "APPLICATION"
    model = "MODEL"
    dataset = "DATASET"


class StorageTypeItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class ResourceNameItem(RootModel[str]):
    root: Annotated[str, Field(max_length=128)]


class ShaItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class RepoTypeItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class RepoNameItem(RootModel[str]):
    root: Annotated[str, Field(max_length=128)]


class RepoUrlItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class GitInfo(RadarBaseModel):
    sha: Annotated[Optional[ShaItem], Field(title="Sha")] = None
    repo_type: Annotated[Optional[RepoTypeItem], Field(title="Repo Type")] = None
    repo_name: Annotated[Optional[RepoNameItem], Field(title="Repo Name")] = None
    repo_url: Annotated[Optional[RepoUrlItem], Field(title="Repo Url")] = None
    dirty: Annotated[Optional[bool], Field(title="Dirty")] = None


class License(RadarBaseModel):
    name: Annotated[str, Field(max_length=128, title="Name")]


class LicenseWrapper(RadarBaseModel):
    license: License


class ModelBaseSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class ModelScanIssue(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    description: Annotated[str, Field(title="Description")]
    operator: Annotated[str, Field(title="Operator")]
    module: Annotated[str, Field(title="Module")]
    source: Annotated[str, Field(title="Source")]
    scanner: Annotated[str, Field(title="Scanner")]


class ModelStatsSchema(RadarBaseModel):
    total_models: Annotated[int, Field(title="Total Models")]
    total_models_without_pipelines: Annotated[
        int, Field(title="Total Models Without Pipelines")
    ]
    models_without_pipelines: Annotated[
        str, Field(max_length=64, title="Models Without Pipelines")
    ]
    total_models_with_policy_failures: Annotated[
        int, Field(title="Total Models With Policy Failures")
    ]
    total_scanned_models: Annotated[int, Field(title="Total Scanned Models")]
    scanned_models_with_issues: Annotated[
        str, Field(max_length=64, title="Scanned Models With Issues")
    ]
    total_unscanned_models: Annotated[int, Field(title="Total Unscanned Models")]


class ModelUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[Optional[NameItem], Field(title="Name")] = None
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class VersionItem3(RootModel[str]):
    root: Annotated[str, Field(max_length=64)]


class PaginationSchema(RadarBaseModel):
    total_items: Annotated[Optional[int], Field(title="Total Items")] = None


class OwnerItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class PipelineCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(max_length=128, title="Name")]
    owner: Annotated[Optional[OwnerItem], Field(title="Owner")] = None
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class ExternalRunUrlItem(RootModel[AnyUrl]):
    root: AnyUrl


class ExternalRunIdItem(RootModel[str]):
    root: Annotated[str, Field(max_length=512)]


class PlatformItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class PipelineRunState(Enum):
    active = "ACTIVE"
    closed = "CLOSED"
    failed = "FAILED"


class AdjacentRun(RootModel[List[Any]]):
    root: Annotated[List[Any], Field(max_length=2, min_length=2)]


class PipelineStatsSchema(RadarBaseModel):
    total_pipeline_runs: Annotated[int, Field(title="Total Pipeline Runs")]
    total_pipelines: Annotated[int, Field(title="Total Pipelines")]


class PipelineUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[Optional[NameItem], Field(title="Name")] = None
    owner: Annotated[Optional[OwnerItem], Field(title="Owner")] = None
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class RemediationTitleItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class RemediationUrlItem(RootModel[AnyUrl]):
    root: AnyUrl


class PolicyEvaluationRequestSchema(RadarBaseModel):
    policy_definition: Annotated[str, Field(title="Policy Definition")]
    policy_input: Annotated[str, Field(title="Policy Input")]


class Error(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class PolicyEvaluationResponseSchema(RadarBaseModel):
    policy_definition: Annotated[str, Field(title="Policy Definition")]
    opa_output: Annotated[Optional[str], Field(title="Opa Output")] = None
    opa_error_code: Annotated[Optional[int], Field(title="Opa Error Code")] = None
    errors: Annotated[Optional[List[Error]], Field(title="Errors")] = None


class PolicyEvaluationResult(Enum):
    success = "SUCCESS"
    failed = "FAILED"
    skipped = "SKIPPED"
    invalid = "INVALID"


class ErrorMessageItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class PolicyEvaluationSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    result: PolicyEvaluationResult
    rule_name: Annotated[str, Field(max_length=128, title="Rule Name")]
    entity_uuid: Annotated[str, Field(max_length=256, title="Entity Uuid")]
    base_entity_uuid: Annotated[str, Field(max_length=256, title="Base Entity Uuid")]
    policy_snapshot: Annotated[str, Field(title="Policy Snapshot")]
    error_message: Annotated[
        Optional[ErrorMessageItem], Field(title="Error Message")
    ] = None


class PolicyEvaluationWithPayloadSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    result: PolicyEvaluationResult
    rule_name: Annotated[str, Field(max_length=128, title="Rule Name")]
    entity_uuid: Annotated[str, Field(max_length=256, title="Entity Uuid")]
    base_entity_uuid: Annotated[str, Field(max_length=256, title="Base Entity Uuid")]
    policy_snapshot: Annotated[str, Field(title="Policy Snapshot")]
    error_message: Annotated[
        Optional[ErrorMessageItem], Field(title="Error Message")
    ] = None
    opa_payload: Annotated[str, Field(title="Opa Payload")]
    opa_output: Annotated[Optional[str], Field(title="Opa Output")] = None


class PolicyStatsSchema(RadarBaseModel):
    total_policies: Annotated[int, Field(title="Total Policies")]
    total_policy_violations: Annotated[int, Field(title="Total Policy Violations")]
    total_failed_evaluations_with_critical_severity: Annotated[
        int, Field(title="Total Failed Evaluations With Critical Severity")
    ]
    total_evaluations_with_critical_severity: Annotated[
        int, Field(title="Total Evaluations With Critical Severity")
    ]


class PolicyStatus(Enum):
    enabled = "ENABLED"
    evaluate = "EVALUATE"
    disabled = "DISABLED"


class RadarEntityAttributes(RadarBaseModel):
    radar_uri: Annotated[str, Field(max_length=256, title="Radar Uri")]


class SearchResultSchema(RadarBaseModel):
    name: Annotated[str, Field(max_length=128, title="Name")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    entity: EntityType
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class SeverityLevel(Enum):
    low = "LOW"
    medium = "MEDIUM"
    high = "HIGH"
    critical = "CRITICAL"


class SkippedSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    total_skipped: Annotated[int, Field(title="Total Skipped")]
    skipped_files: Annotated[List[str], Field(title="Skipped Files")]


class TotalIssuesBySeveritySchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    low: Annotated[int, Field(alias="LOW", title="Low")]
    medium: Annotated[int, Field(alias="MEDIUM", title="Medium")]
    high: Annotated[int, Field(alias="HIGH", title="High")]
    critical: Annotated[int, Field(alias="CRITICAL", title="Critical")]


class TotalIssuesSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    total_issues_by_severity: TotalIssuesBySeveritySchema


class ValidationError(RadarBaseModel):
    loc: Annotated[List[Union[str, int]], Field(title="Location")]
    msg: Annotated[str, Field(title="Message")]
    type: Annotated[str, Field(title="Error Type")]


class VersionStrategy(Enum):
    s3 = "S3"
    custom = "CUSTOM"


class CveIdItem(RootModel[str]):
    root: Annotated[str, Field(max_length=256)]


class Vulnerability(RadarBaseModel):
    summary: Annotated[str, Field(max_length=256, title="Summary")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    severity: CVESeverity
    cve_id: Annotated[Optional[CveIdItem], Field(title="Cve Id")] = None


class DatasetSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    version_count: Annotated[Optional[int], Field(title="Version Count")] = 0
    latest_version: Optional[DatasetVersionWithDatasetSchema] = None


class DatasetVersionListSchema(RadarBaseModel):
    pagination: PaginationSchema
    versions: Annotated[List[DatasetVersionWithDatasetSchema], Field(title="Versions")]


class Dependency(RadarBaseModel):
    name: Annotated[str, Field(max_length=128, title="Name")]
    version: Annotated[Optional[VersionItem1], Field(title="Version")] = None
    author: Annotated[Optional[AuthorItem], Field(title="Author")] = None
    purl: Annotated[Optional[PurlItem], Field(title="Purl")] = None
    licenses: Annotated[Optional[List[LicenseWrapper]], Field(title="Licenses")] = None
    sightline: Annotated[Optional[List[Vulnerability]], Field(title="Sightline")] = None


class File(RadarBaseModel):
    uri: Annotated[str, Field(max_length=256, title="Uri")]
    storage_type: Annotated[Optional[StorageTypeItem], Field(title="Storage Type")] = (
        None
    )
    resource_name: Annotated[
        Optional[ResourceNameItem], Field(title="Resource Name")
    ] = None
    size: Annotated[Optional[int], Field(title="Size")] = None
    version: Annotated[Optional[VersionItem1], Field(title="Version")] = None
    version_strategy: Optional[VersionStrategy] = None


class FileListSchema(RadarBaseModel):
    pagination: PaginationSchema
    files: Annotated[Optional[List[File]], Field(title="Files")] = None


class HTTPValidationError(RadarBaseModel):
    detail: Annotated[Optional[List[ValidationError]], Field(title="Detail")] = None


class IssuesBySeveritySchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    low: Annotated[Optional[List[ModelScanIssue]], Field(alias="LOW", title="Low")] = []
    medium: Annotated[
        Optional[List[ModelScanIssue]], Field(alias="MEDIUM", title="Medium")
    ] = []
    high: Annotated[
        Optional[List[ModelScanIssue]], Field(alias="HIGH", title="High")
    ] = []
    critical: Annotated[
        Optional[List[ModelScanIssue]], Field(alias="CRITICAL", title="Critical")
    ] = []


class ModelScanResultSchemaInput(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    modelscan_version: Annotated[str, Field(title="Modelscan Version")]
    timestamp: Annotated[datetime, Field(title="Timestamp")]
    input_path: Annotated[str, Field(title="Input Path")]
    total_issues: Annotated[int, Field(title="Total Issues")]
    summary: TotalIssuesSchema
    issues_by_severity: IssuesBySeveritySchema
    errors: Annotated[List[str], Field(title="Errors")]
    skipped: SkippedSchema


class ModelScanResultSchemaOutput(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    modelscan_version: Annotated[str, Field(title="Modelscan Version")]
    timestamp: Annotated[datetime, Field(title="Timestamp")]
    input_path: Annotated[str, Field(title="Input Path")]
    total_issues: Annotated[int, Field(title="Total Issues")]
    summary: TotalIssuesSchema
    issues_by_severity: IssuesBySeveritySchema
    errors: Annotated[List[str], Field(title="Errors")]
    skipped: SkippedSchema


class ModelVersionCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(max_length=128, title="Name")]
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    version: Annotated[Optional[VersionItem3], Field(title="Version")] = None
    files: Annotated[Optional[List[File]], Field(title="Files")] = None
    properties: Annotated[Optional[Dict[str, str]], Field(title="Properties")] = None
    model_scan_result: Optional[ModelScanResultSchemaInput] = None


class PipelineRunCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    external_run_url: Annotated[
        Optional[ExternalRunUrlItem], Field(title="External Run Url")
    ] = None
    external_run_id: Annotated[
        Optional[ExternalRunIdItem], Field(title="External Run Id")
    ] = None
    platform: Annotated[Optional[PlatformItem], Field(title="Platform")] = None
    dependencies: Annotated[Optional[List[Dependency]], Field(title="Dependencies")] = (
        None
    )
    git_info: Optional[GitInfo] = None
    run_at: Annotated[datetime, Field(title="Run At")]


class PipelineRunSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    external_run_url: Annotated[
        Optional[ExternalRunUrlItem], Field(title="External Run Url")
    ] = None
    external_run_id: Annotated[
        Optional[ExternalRunIdItem], Field(title="External Run Id")
    ] = None
    platform: Annotated[Optional[PlatformItem], Field(title="Platform")] = None
    model_count: Annotated[Optional[int], Field(title="Model Count")] = 0
    run_at: Annotated[datetime, Field(title="Run At")]
    completed_at: Annotated[Optional[datetime], Field(title="Completed At")] = None
    status: PipelineRunState
    git_info: Optional[GitInfo] = None


class PipelineRunUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    dependencies: Annotated[Optional[List[Dependency]], Field(title="Dependencies")] = (
        None
    )
    status: Optional[PipelineRunState] = None
    completed_at: Annotated[Optional[datetime], Field(title="Completed At")] = None


class PipelineSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    owner: Annotated[Optional[OwnerItem], Field(title="Owner")] = None
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    run_count: Annotated[Optional[int], Field(title="Run Count")] = 0
    latest_run: Optional[PipelineRunSchema] = None
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class PolicyCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(max_length=128, title="Name")]
    status: PolicyStatus
    definition: Annotated[str, Field(title="Definition")]
    scope: EntityType
    severity: SeverityLevel
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    remediation_title: Annotated[
        Optional[RemediationTitleItem], Field(title="Remediation Title")
    ] = None
    remediation_url: Annotated[
        Optional[RemediationUrlItem], Field(title="Remediation Url")
    ] = None


class PolicyEvaluationListSchema(RadarBaseModel):
    pagination: PaginationSchema
    evaluations: Annotated[List[PolicyEvaluationSchema], Field(title="Evaluations")]


class PolicySchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    status: PolicyStatus
    definition: Annotated[str, Field(title="Definition")]
    scope: EntityType
    severity: SeverityLevel
    total_violations: Annotated[int, Field(title="Total Violations")]
    total_evaluations: Annotated[int, Field(title="Total Evaluations")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    latest_evaluation: Optional[PolicyEvaluationSchema] = None
    remediation_title: Annotated[
        Optional[RemediationTitleItem], Field(title="Remediation Title")
    ] = None
    remediation_url: Annotated[
        Optional[RemediationUrlItem], Field(title="Remediation Url")
    ] = None


class PolicyUpdateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[Optional[NameItem], Field(title="Name")] = None
    status: Optional[PolicyStatus] = None
    definition: Annotated[Optional[str], Field(title="Definition")] = None
    severity: Optional[SeverityLevel] = None
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    remediation_title: Annotated[
        Optional[RemediationTitleItem], Field(title="Remediation Title")
    ] = None
    remediation_url: Annotated[
        Optional[RemediationUrlItem], Field(title="Remediation Url")
    ] = None


class SearchListSchema(RadarBaseModel):
    pagination: PaginationSchema
    results: Annotated[Optional[List[SearchResultSchema]], Field(title="Results")] = []


class DatasetListSchema(RadarBaseModel):
    pagination: PaginationSchema
    datasets: Annotated[List[DatasetSchema], Field(title="Datasets")]


class DatasetVersionCreateRequestSchema(RadarBaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    name: Annotated[str, Field(max_length=128, title="Name")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    version: Annotated[Optional[VersionItem], Field(title="Version")] = None
    files: Annotated[Optional[List[File]], Field(title="Files")] = None


class PipelineListSchema(RadarBaseModel):
    pagination: PaginationSchema
    pipelines: Annotated[List[PipelineSchema], Field(title="Pipelines")]


class PipelineRunListSchema(RadarBaseModel):
    pagination: PaginationSchema
    pipeline_runs: Annotated[List[PipelineRunSchema], Field(title="Pipeline Runs")]


class PipelineRunWithPipelineSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    external_run_url: Annotated[
        Optional[ExternalRunUrlItem], Field(title="External Run Url")
    ] = None
    external_run_id: Annotated[
        Optional[ExternalRunIdItem], Field(title="External Run Id")
    ] = None
    platform: Annotated[Optional[PlatformItem], Field(title="Platform")] = None
    model_count: Annotated[Optional[int], Field(title="Model Count")] = 0
    run_at: Annotated[datetime, Field(title="Run At")]
    completed_at: Annotated[Optional[datetime], Field(title="Completed At")] = None
    status: PipelineRunState
    git_info: Optional[GitInfo] = None
    pipeline: PipelineSchema
    adjacent_runs: Annotated[
        Optional[List[AdjacentRun]], Field(title="Adjacent Runs")
    ] = None


class PolicyListSchema(RadarBaseModel):
    pagination: PaginationSchema
    policies: Annotated[List[PolicySchema], Field(title="Policies")]


class ArtifactPipelineRunsSchema(RadarBaseModel):
    primary_run: Optional[PipelineRunWithPipelineSchema] = None
    secondary_runs: Annotated[
        Optional[List[PipelineRunWithPipelineSchema]], Field(title="Secondary Runs")
    ] = []


class PipelineRunAssocation(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    pipeline_run: PipelineRunWithPipelineSchema


class ModelVersionDatasetListSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    file_count: Annotated[int, Field(title="File Count")]
    version: Annotated[str, Field(max_length=64, title="Version")]
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    properties: Annotated[Optional[Dict[str, str]], Field(title="Properties")] = None
    evaluation_status: Optional[PolicyEvaluationResult] = None
    primary_pipeline_run_association: Optional[PipelineRunAssocation] = None
    model_scan_result: Optional[ModelScanResultSchemaOutput] = None
    model: ModelBaseSchema
    bom_available: Annotated[bool, Field(title="Bom Available")]


class ModelVersionSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    file_count: Annotated[int, Field(title="File Count")]
    version: Annotated[str, Field(max_length=64, title="Version")]
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    properties: Annotated[Optional[Dict[str, str]], Field(title="Properties")] = None
    evaluation_status: Optional[PolicyEvaluationResult] = None
    primary_pipeline_run_association: Optional[PipelineRunAssocation] = None
    model_scan_result: Optional[ModelScanResultSchemaOutput] = None


class ModelVersionWithModelSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    file_count: Annotated[int, Field(title="File Count")]
    version: Annotated[str, Field(max_length=64, title="Version")]
    external_uri: Annotated[str, Field(max_length=256, title="External Uri")]
    properties: Annotated[Optional[Dict[str, str]], Field(title="Properties")] = None
    evaluation_status: Optional[PolicyEvaluationResult] = None
    primary_pipeline_run_association: Optional[PipelineRunAssocation] = None
    model_scan_result: Optional[ModelScanResultSchemaOutput] = None
    model: ModelBaseSchema
    bom_available: Annotated[bool, Field(title="Bom Available")]
    direction: Optional[ArtifactDirection] = None
    files: Annotated[Optional[List[File]], Field(title="Files")] = None


class DeploymentSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    environment: Annotated[str, Field(max_length=256, title="Environment")]
    build_time: Annotated[datetime, Field(title="Build Time")]
    application_url: Annotated[
        Optional[ApplicationUrlItem], Field(title="Application Url")
    ] = None
    build_url: Annotated[Optional[BuildUrlItem], Field(title="Build Url")] = None
    model_versions: Annotated[
        Optional[List[ModelVersionWithModelSchema]], Field(title="Model Versions")
    ] = None
    application_version: Annotated[
        Optional[ApplicationVersionItem], Field(title="Application Version")
    ] = None


class ModelSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None
    version_count: Annotated[int, Field(title="Version Count")]
    latest_version: Optional[ModelVersionSchema] = None


class ModelVersionListSchema(RadarBaseModel):
    pagination: PaginationSchema
    versions: Annotated[List[ModelVersionWithModelSchema], Field(title="Versions")]


class ApplicationSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    name: Annotated[str, Field(max_length=128, title="Name")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    external_url: Annotated[Optional[ExternalUrlItem], Field(title="External Url")] = (
        None
    )
    total_deployment_count: Annotated[
        Optional[int], Field(title="Total Deployment Count")
    ] = 0
    latest_deployment: Optional[DeploymentSchema] = None
    description: Annotated[Optional[DescriptionItem], Field(title="Description")] = None


class DeploymentWithApplicationSchema(RadarBaseModel):
    created_at: Annotated[datetime, Field(title="Created At")]
    updated_at: Annotated[datetime, Field(title="Updated At")]
    uuid: Annotated[str, Field(max_length=256, title="Uuid")]
    environment: Annotated[str, Field(max_length=256, title="Environment")]
    build_time: Annotated[datetime, Field(title="Build Time")]
    application_url: Annotated[
        Optional[ApplicationUrlItem], Field(title="Application Url")
    ] = None
    build_url: Annotated[Optional[BuildUrlItem], Field(title="Build Url")] = None
    model_versions: Annotated[
        Optional[List[ModelVersionWithModelSchema]], Field(title="Model Versions")
    ] = None
    application: ApplicationSchema


class DeploymentsListSchema(RadarBaseModel):
    pagination: PaginationSchema
    deployments: Annotated[
        List[DeploymentWithApplicationSchema], Field(title="Deployments")
    ]


class ModelListSchema(RadarBaseModel):
    pagination: PaginationSchema
    models: Annotated[List[ModelSchema], Field(title="Models")]


class ApplicationListSchema(RadarBaseModel):
    pagination: PaginationSchema
    applications: Annotated[List[ApplicationSchema], Field(title="Applications")]



import os
import logging
from datetime import datetime

from typing import Dict, Union, Optional, List

from airadar import typing as radar_types
from airadar.typing import RadarResponse

from airadar.utils import get_dependencies_from_environment
from airadar.api.api_client import RadarAPIClient
from airadar.tracking.model import PipelineRunState, Dependency, ExternalUrlItem
from airadar.utils import get_files_and_name_for_uri

logger = logging.getLogger(__name__)


class Radar:
    def __init__(self) -> None:
        self.radar_api_client = RadarAPIClient()
        if self.radar_api_client.is_authorized():
            self.disable_data_collection = False
        else:
            logger.error(
                "Authorization failed, disabling data collection for Radar instance"
            )
            self.disable_data_collection = True

    def log_artifact(
        self,
        uri: str,
        name: Union[str, None] = None,
        artifact_type: radar_types.AnyArtifact = radar_types.Model,
        description: Optional[str] = None,
        version: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> RadarResponse:
        """
        Log a collection of files from a given path as a model without pipeline.

        The base method for [log_dataset][airadar.radar.Radar.log_dataset] and
        [log_model][airadar.radar.Radar.log_model]

        Arguments:
            uri: URI of the artifact, without any schema prefix a local path is assumed. You can also specify glob pattern on local filesystem or cloud storage
                If a directory path is specified, all its contents will be recursively associated with this model.
            name: Human-readable name of the artifact.
                If None, the artifact directory path for multiple files or file name for the single file is used.
            artifact_type: Should be [Model][airadar.typing.Model] as we currently only support logging models without a pipeline.
            description: Human-readable description of the artifact
            version: A version string relevant to your environment. If no version string is applied. Radar will assign an auto-incremented numeric version
                to the artifact.
            properties: A dictionary of arbitrary key/value pairs that the user can provide for their own use cases.

        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the model or dataset version

        Examples:

        To log a dataset that was consumed as input by a given pipeline stage:
        ```
            import radar
            radar.log_artifact(
                uri="s3://bucket_name/training_data",
                name="flower_training_data",
                artifact_type=radar_types.Dataset,
            )
        ```
        """
        if artifact_type != radar_types.Model:
            error_msg = f"Only models can be logged without a pipeline context"
            logger.error(error_msg)
            return RadarResponse(error_msg=error_msg)

        if not uri or not isinstance(uri, str):
            error_msg = f"Artifact uri {uri} was empty or not of accepted type: str"
            logger.warning(error_msg)
            return RadarResponse(error_msg=error_msg)

        files, inferred_name = get_files_and_name_for_uri(uri)

        if not name:
            if not inferred_name:
                error_msg = f"Name for {uri} could not be inferred, provide a name for artifact to be logged"
                logger.warning(error_msg)
                return RadarResponse(error_msg=error_msg)

            name = inferred_name

        radar_response = self.radar_api_client.post_model(
            name=name,
            uri=uri,
            files=files,
            description=description,
            version=version,
            properties=properties,
        )

        return radar_response

    def log_model(
        self,
        uri: str,
        name: Union[str, None] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> RadarResponse:
        """
        Log a model with associated files.

        Arguments:
            uri: URI of the model preferably belonging to a model with your model registry, without any schema prefix a local path is assumed.
                You can also specify glob pattern on local filesystem or cloud storage
                If a directory path is specified, all its contents will be recursively associated with this model.
            name: Human-readable name of the model this file is a part of.
                If None, the absolute filepath (for single file) or directory name (for directory path) is used.
            description: Human-readable description of the model. This is displayed on the Radar's model pages.
            version: A version string relevant to your environment. If no version string is applied. Radar will assign an auto-incremented numeric version
                to the model.
            properties: A dictionary of arbitrary key/value pairs that the user can provide for their own use cases.

        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the model version
        """

        return self.log_artifact(
            uri,
            name,
            radar_types.Model,
            description,
            version,
            properties,
        )

    def log_dataset(
        self,
        uri: str,
        name: Union[str, None] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
    ) -> RadarResponse:
        """
        Log a dataset with associated files.

        Arguments:
            path: URI of the dataset, without any schema prefix a local path is assumed.
                You can also specify glob pattern on local filesystem or cloud storage
                If a directory path is specified, all its contents will be recursively associated with this dataset
            name: Human-readable name of the dataset this file is a part of. If None, the absolute filepath is used.
            description: Human-readable description of the artifact
            version: Either a version string to be applied to all files associated with the logged dataset or a function which accepts a
                file's absolute path and returns a version string. If None, the SDK will attempt to infer the version for S3 resources.

        A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the dataset version
        """
        return self.log_artifact(
            uri=uri,
            name=name,
            artifact_type=radar_types.Dataset,
            description=description,
            version=version,
        )

    def log_deployment(
        self,
        application_id: str,
        application_version: str,
        environment: str,
        application_url: Optional[str] = None,
        build_url: Optional[str] = None,
        radar_uris: Optional[List[str]] = None,
        build_time: datetime = datetime.now(),
    ) -> RadarResponse:
        """
        Creates a new deployment against an application.

        Deployments are used to monitor deployed models within applications running in a user's organization.
        These applications might include, but aren't limited to, inference applications that
        serve the model to other users or software.

        Ideally, you should integrate this call into your CI/CD pipeline responsible for deploying the model.
        The CI/CD pipelines can provide many of the arguments to this function.
        This information can then allow Radar to trace back the provenance of the model within your application

        Note: application_id is Radar specific unique identifier obtained from onboarding the
        application on Radar web console.

        Arguments:
            application_id: The uuid of the application obtained from Radar's console.
            application_version: The version of the application within users' environment.
            environment: The name of the environment e.g. production, staging, dev etc.
            application_url: The URI to identify the application.
            build_url: This is typically the URI of the CI/CD pipeline that is deploying the call.
            radar_uris: The URI of the model in the form of radar://model/<model-name>/<model-version>.
            build_time: The date and time at which the application was built e.g. injected by your CI/CD pipeline.

        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the deployment

        Example:
            ```
            import airadar
            airadar.log_deployment(
                application_id="..",
                application_version="1.0.5",
                environment="production",
                application_uri="https://example.com/api/v1/service",
                build_uri="https://github.com/org/repo/actions/run/4",
                radar_uris=["radar://model/cost-predictions/latest"],
            )
            ```
        """
        if self.disable_data_collection:
            return RadarResponse(error_msg="Client data collection disabled")

        return self.radar_api_client.post_deployment(
            application_id=application_id,
            application_version=application_version,
            environment=environment,
            application_url=application_url,
            build_url=build_url,
            radar_uris=radar_uris,
            build_time=build_time,
        )


class RadarPipeline(Radar):
    def __init__(
        self,
        pipeline_id: Optional[str] = os.environ.get("AIRADAR_PIPELINE_ID"),
        editable: bool = False,
        external_run_id: Optional[str] = None,
        external_run_uri: Optional[str] = None,
        dependencies: Optional[List[Dependency]] = None,
        pipeline_name: Optional[str] = None,
        external_url: Optional[ExternalUrlItem] = None,
    ) -> None:
        super().__init__()

        self.pipeline_id = pipeline_id
        self.pipeline_name = pipeline_name
        self.pipeline_run_id: Optional[str] = None
        self.external_run_id = external_run_id
        self.external_run_uri = external_run_uri
        self.dependencies = dependencies
        self.props: Dict[str, Union[str, int, float, bool]] = {}
        self.editable = editable
        self.external_url = external_url

        if not self.pipeline_id and not self.pipeline_name:
            error_msg = "A Radar pipeline ID or name is required to initiate data collection session. Radar will not be able to collect any data"
            logger.critical(error_msg)
            self.disable_data_collection = True

    def start_run(self) -> RadarResponse:
        if self.disable_data_collection:
            return RadarResponse(error_msg="Data collection was disabled")

        if not self.pipeline_id:
            logger.info("No pipeline ID found, creating a new pipeline")

            if not self.pipeline_name:
                error_msg = "Pipeline name is required to create a new pipeline. Data collection will be disabled."
                logger.error(error_msg)
                return RadarResponse(error_msg=error_msg)

            response = self.radar_api_client.post_pipeline(
                pipeline_name=self.pipeline_name, external_url=self.external_url
            )

            if not response.is_success():
                logger.info(
                    "We couldn't create a new pipeline, disabling data collection"
                )
                self.disable_data_collection = True
                return response

            self.pipeline_id = response.entity_uuid
            logger.info("Obtained pipeline ID: %s", self.pipeline_id)

        # We should have pipeline id at this point either through user provided or obtained with name.
        if not self.pipeline_id:
            error_msg = "Pipeline ID is required to start a pipeline run. Data collection will be disabled."
            logger.error(error_msg)
            self.disable_data_collection = True
            return RadarResponse(error_msg=error_msg)

        if not self.dependencies:
            dependency_report = get_dependencies_from_environment()
        else:
            dependency_report = self.dependencies

        response = self.radar_api_client.post_pipeline_run(
            pipeline_id=self.pipeline_id,
            external_run_id=self.external_run_id,
            run_uri=self.external_run_uri,
            dependency_report=dependency_report,
        )

        if not response.is_success():
            self.disable_data_collection = True
            return response

        self.pipeline_run_id = response.entity_uuid

        return response

    def finish_run(self, failure: bool = False) -> RadarResponse:
        if not self.pipeline_run_id or self.disable_data_collection:
            return RadarResponse(error_msg="Finish run called on a non-active pipeline")

        completed_at = datetime.now()
        if failure:
            status = PipelineRunState.failed
        elif self.editable:
            status = PipelineRunState.active
        else:
            status = PipelineRunState.closed

        res = self.radar_api_client.patch_pipeline_run(
            status, self.pipeline_run_id, completed_at
        )

        self.disable_data_collection = True
        return res

    def log_artifact(
        self,
        uri: str,
        name: Union[str, None] = None,
        artifact_type: radar_types.AnyArtifact = radar_types.Model,
        description: Optional[str] = None,
        version: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> RadarResponse:
        """
        Log a collection of files from a given path as either model or dataset within a Radar pipeline.

        The base method for [log_dataset][airadar.radar.RadarPipeline.log_dataset]
        and [log_model][airadar.radar.RadarPipeline.log_model]

        Arguments:
            uri: URI of the artifact, without any schema prefix a local path is assumed. You can also specify glob pattern on local filesystem or cloud storage
                If a directory path is specified, all its contents will be recursively associated with this model.
            name: Human-readable name of the artifact.
                If None, the artifact directory path for multiple files or file name for the single file is used.
            artifact_type: Either of [Model][airadar.typing.Model] or [Dataset][airadar.typing.Dataset]. Model is the default.
            description: Human-readable description of the artifact
            version: A version string relevant to your environment. If no version string is applied. Radar will assign an auto-incremented numeric version
                to the artifact.


        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the model or dataset version

        Examples:

        To log a dataset that was consumed as input by a given pipeline stage:
        ```
            import radar
            radar.log_artifact(
                uri="s3://bucket_name/training_data.tar.gz",
                artifact_name="flower_training_data",
                artifact_type=radar_types.Dataset,
            )
        ```
        """

        if self.disable_data_collection:
            return RadarResponse(error_msg="Client data collection disabled")

        if not self.pipeline_run_id:
            error_msg = f"Active pipeline run is required to log an artifact. Did you call start_run()?"
            logger.error(error_msg)
            return RadarResponse(error_msg=error_msg)

        if artifact_type not in [radar_types.Dataset, radar_types.Model]:
            error_msg = f"Cannot log {name} of unsupported type {artifact_type}"
            logger.error(error_msg)
            return RadarResponse(error_msg=error_msg)

        if not uri or not isinstance(uri, str):
            error_msg = (
                f"Artifact uri {uri} was either empty or not an accepted type str"
            )
            logger.warning(error_msg)
            return RadarResponse(error_msg=error_msg)

        files, inferred_name = get_files_and_name_for_uri(uri)

        if not name:
            if not inferred_name:
                error_msg = f"Name for {uri} could not be inferred, provide a name for artifact to be logged"
                logger.warning(error_msg)
                return RadarResponse(error_msg=error_msg)

            name = inferred_name

        if artifact_type == radar_types.Model:
            radar_response = self.radar_api_client.post_model(
                name=name,
                uri=uri,
                files=files,
                version=version,
                description=description,
                pipeline_run_id=self.pipeline_run_id,
                properties=properties,
            )
        elif artifact_type == radar_types.Dataset:
            if not self.pipeline_run_id:
                error_msg = f"Datasets can only be logged with an active pipeline run"
                logger.error(error_msg)

                return RadarResponse(error_msg=error_msg)

            radar_response = self.radar_api_client.post_dataset(
                name=name,
                uri=uri,
                files=files,
                pipeline_run_id=self.pipeline_run_id,
                description=description,
                version=version,
            )

        if not radar_response.is_success():
            self.disable_data_collection = True

        return radar_response

    def log_deployment(
        self,
        application_id: str,
        application_version: str,
        environment: str,
        application_uri: Optional[str] = None,
        build_uri: Optional[str] = None,
        radar_uris: Optional[List[str]] = None,
        build_time: datetime = datetime.now(),
    ) -> RadarResponse:
        """
        Creates a new deployment against an application.

        Deployments are used to monitor deployed models within applications running in a user's organization.
        These applications might include, but aren't limited to, inference applications that
        serve the model to other users or software.

        Ideally, you should integrate this call into your CI/CD pipeline responsible for deploying the model.
        The CI/CD pipelines can provide many of the arguments to this function.
        This information can then allow Radar to trace back the provenance of the model within your application

        Note: application_id is Radar specific unique identifier obtained from onboarding the
        application on Radar web console.

        Arguments:
            application_id: The uuid of the application obtained from Radar's console.
            application_version: The version of the application within users' environment.
            environment: The name of the environment e.g. production, staging, dev etc.
            application_uri: The URI to identify the application.
            build_uri: This is typically the URI of the CI/CD pipeline that is deploying the call.
            radar_uris: The URI of the model in the form of radar://model/<model-name>/<model-version>.
            build_time: The date and time at which the application was built e.g. injected by your CI/CD pipeline.

        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the deployment

        Example:
            ```
            import airadar
            airadar.log_deployment(
                application_id="..",
                application_version="1.0.5",
                environment="production",
                application_uri="https://example.com/api/v1/service",
                build_uri="https://github.com/org/repo/actions/run/4",
                radar_uris=["radar://model/cost-predictions/latest"],
            )
            ```
        """
        return RadarResponse(
            error_msg="log_deployment is only possible outside pipeline context"
        )

    def log_input_model(
        self,
        radar_uri: str,
    ) -> RadarResponse:
        """
        Log an existing Radar model to current pipeline run. This model will be associated as INPUT (base model) used by the pipeline.


        Arguments:
            radar_uri: Radar specific URI of the model. It takes the following format: radar://model/<model-name>/<model-version>

        Returns:
            A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
                Radar API call to log the model.

        Examples:

        To associate an model that was previously logged with Radar, find its URI from the Radar dashboard and log it as input:
        ```
            import airadar
            with airadar.active_run(pipeline_id="..") as radar:
                radar.log_input_model(
                    radar_uri="radar://model/credit risk score/v1",
                )
        ```
        """

        if self.disable_data_collection:
            return RadarResponse(error_msg="Client data collection disabled")

        if not self.pipeline_run_id:
            error_msg = f"Active pipeline run is required to log an input model. Did you call start_run()?"
            logger.error(error_msg)
            return RadarResponse(error_msg=error_msg)

        return self.radar_api_client.post_input_model(
            radar_uri=radar_uri, pipeline_run_id=self.pipeline_run_id
        )

    def is_closed(self) -> bool:
        return self.disable_data_collection



import requests
import logging
import jwt
from typing import Optional
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


def get_ccf_access_token(
    auth_tenant_url: str,
    auth_client_id: str,
    auth_client_secret: str,
    auth_client_scope: str,
) -> Optional[str]:
    """
    Retrieves a JWT access token using client credentials flow.

    Arguments:
        auth_tenant_url: The URL for your auth provider
        auth_client_id: The client ID
        auth_client_secret: The client secret
        auth_client_scope: [Optional] The permissions scope for the access token
    """

    payload = {
        "grant_type": "client_credentials",
        "client_id": auth_client_id,
        "client_secret": auth_client_secret,
        "roles": "api://radar-api/sdk",  # This adds the correct role that the api will verify
    }

    return _get_access_token(auth_tenant_url, payload)


def get_private_key_access_token(
    tenant_url: str,
    client_id: str,
    private_key: str,
    key_id: str,
    client_scope: str,
) -> Optional[str]:
    """
    Retrieves an API Access token based on a JWT signed with a user-provided private key.

    Arguments:
        tenant_url: The URL for your auth provider
        client_id: The client ID
        private_key: The client private key to sign the JWT
        key_id: The key ID to identify the private key
        client_scope: The permissions scope for the access token
    """

    # Set JWT expiration time (e.g., 1 hour from now)
    expiration_time = datetime.utcnow() + timedelta(hours=1)

    # Prepare JWT payload
    jwt_payload = {
        "iss": client_id,  # Client ID
        "sub": client_id,  # Client ID
        "aud": tenant_url,  # Token URL
        "exp": expiration_time,  # Expiration time
    }

    # Generate JWT token signed with private key
    jwt_token = jwt.encode(
        jwt_payload,
        private_key,
        algorithm="RS256",
        headers={"kid": key_id},
    )

    # Make a POST request with JWT token
    payload = {
        "grant_type": "client_credentials",
        "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
        "client_assertion": jwt_token,
        "scope": client_scope,
        "roles": "api://radar-api/sdk",
    }
    return _get_access_token(tenant_url, payload)


def _get_access_token(tenant_url: str, payload: dict[str, str]) -> Optional[str]:
    try:
        response = requests.post(
            tenant_url,
            data=payload,
            timeout=2,
        )

        resp = response.json().get("access_token")
        if resp and type(resp) is str:
            return resp
    except requests.exceptions.HTTPError as e:
        if e.response is not None:
            logger.error("Error with request: %s", e.response.status_code)
            logger.error(e.response.text)
        else:
            logger.error("Unknown HTTPError with request to %s", tenant_url)
    except requests.exceptions.JSONDecodeError as e:
        logger.error(
            "Response was not valid JSON for request to %s: %s", tenant_url, str(e)
        )
    except requests.exceptions.Timeout as e:
        logger.error("Request timed out for request to %s: %s", tenant_url, str(e))
    except Exception as e:
        logger.error("Unexpected error during request to %s: %s", tenant_url, str(e))

    return None



import os
import logging
from enum import Enum
from dotenv import dotenv_values
from typing import Dict, Union


logger = logging.getLogger(__name__)


class ConfigKeys(str, Enum):
    AIRADAR_API_SERVER = "AIRADAR_API_SERVER"
    AIRADAR_API_VERSION = "AIRADAR_API_VERSION"
    AIRADAR_AUTH_CLIENT_ID = "AIRADAR_AUTH_CLIENT_ID"
    AIRADAR_AUTH_CLIENT_SECRET = "AIRADAR_AUTH_CLIENT_SECRET"  # nosec
    AIRADAR_AUTH_CLIENT_SCOPE = "AIRADAR_AUTH_CLIENT_SCOPE"
    AIRADAR_AUTH_CLIENT_PRIVATE_KEY = "AIRADAR_AUTH_CLIENT_PRIVATE_KEY"
    AIRADAR_AUTH_CLIENT_KID = "AIRADAR_AUTH_CLIENT_KID"
    AIRADAR_AUTH_TENANT_URL = "AIRADAR_AUTH_TENANT_URL"
    AIRADAR_LOG_LEVEL = "AIRADAR_LOG_LEVEL"


class Configurations:
    def __init__(self) -> None:
        self._settings_dict: Dict[ConfigKeys, Union[str, int, None]] = {
            ConfigKeys.AIRADAR_API_VERSION: "v1",
            ConfigKeys.AIRADAR_AUTH_CLIENT_SCOPE: "radar-api",
        }
        self.required_envs_keys = [
            ConfigKeys.AIRADAR_AUTH_TENANT_URL,
            ConfigKeys.AIRADAR_AUTH_CLIENT_ID,
            ConfigKeys.AIRADAR_API_SERVER,
        ]
        self._load_values()

    def _load_values(self) -> None:
        config = {**dotenv_values(".env"), **dotenv_values(".env.local"), **os.environ}

        for key in ConfigKeys:
            self._settings_dict[key] = config.get(key, self._settings_dict.get(key))

        for required_envs_key in self.required_envs_keys:
            if not self._settings_dict.get(required_envs_key):
                logger.critical(
                    f"""
                    {required_envs_key.name} is missing. Logged artifacts won't persist to the server.
                    Set an environment variables for {", ".join([k for k in self.required_envs_keys])}. 
                """
                )
                break

    def get_value(
        self, key: ConfigKeys, default: Union[str, int, None] = None
    ) -> Union[str, int, None]:
        return self._settings_dict.get(key, default)


airadar_configs = Configurations()



import os
from mlflow.tracking.context.abstract_context import RunContextProvider

import airadar
from airadar.tracking.mlflow import start_mlflow_tracking
from typing import Dict, Any


class AIRadarRunContextProvider(RunContextProvider):  # type: ignore
    """Provide AIRadar context through plugin system"""

    def __init__(self) -> None:
        self._activated = False

    def in_context(self) -> bool:
        # MLFlow can trigger this plugin multiple times as part of one run
        # To avoid duplicate artifact logging we make sure our plugin only runs once
        if not self._activated:
            start_mlflow_tracking()
            self._activated = True
        return True

    def tags(self) -> Dict[Any, Any]:
        airadar_pipeline_id = os.environ.get("AIRADAR_PIPELINE_ID", "")
        airadar_console_url = os.environ.get(
            "AIRADAR_CONSOLE_URL", "https://dbx-demo.radar.protectai.com/pipelines"
        )
        return {
            "airadar.pipeline_uri": f"{airadar_console_url}/{airadar_pipeline_id}",
            "airadar.version": f"{airadar.__version__}",
        }



import functools
import inspect
import logging
import os

from typing import Optional, Any, Dict, Callable
from docstring_parser import parse
from airadar import radar_types
from airadar.radar import RadarPipeline
from airadar.typing.types_utils import (
    is_artifact,
    is_radar_type,
    is_dataset_artifact,
    is_model_artifact,
)


logger = logging.getLogger(__name__)


def pipeline(
    func: Optional[Callable[..., Any]] = None,
    pipeline_id: Optional[str] = os.environ.get("AIRADAR_PIPELINE_ID"),
    external_run_id: Optional[str] = None,
    external_run_uri: Optional[str] = None,
    editable: bool = False,
    return_var_name: Optional[str] = None,
    argument_type_dictionary: Optional[Dict[str, radar_types.BaseType]] = None,
) -> Callable[..., Any]:
    """
    Pipeline function decorator that allow Radar to use the arguments and return value of the function to capture artifact paths.

    Within a decorated function, Radar looks for variables annotated with [radar_types][airadar.typing], using the values
     of these variables as artifact paths. Alternatively to type hints, you can pass a dictionary mapping
     variables names to types. See the examples below.

    Arguments:
        func: Decorated function
        pipeline_id: The uuid of the pipeline obtained from Radar's web dashboard.
        external_run_id: ID for a pipeline run provided by a 3rd party tool (e.g. MLflow or Kubeflow).
        external_run_uri: URI identifying this pipeline run from the platform it ran on.
        editable: To keep pipeline open to log additional artifacts to it, set this to True.
        argument_type_dictionary: You can mark arguments to your decorated function using type dictionary.
        return_var_name: Name of the return variable is used by Radar to give a meaningful name to the artifact returned by decorated function.
    Returns:
        A function with the same signature as the decorated one, but is wrapped with Radar instrumentation

    Examples:

        ```python
        @pipeline
        def training(daily_sales_record: radar_types.Dataset[Path]):
            '''
            :param source: Daily sales record
            '''
            ...
        ```
        Radar use variable names as artifact names and use function docstring (if provided) to
        extract description of the variable. Type hints are used to mark a path as either dataset
        or model.


        In the above example, Radar will log the path to data as input dataset using
        `daily_sales_record` as its name and  `Daily sales record` from docstring as the
        description of this dataset.

        You can also tell Radar about the arguments to the function via an argument dictionary instead
        of using type hints

        ```python
        argument_dict = {
            'source': radar.Dataset,
            'return': radar.Dataset
        }

        @pipeline(argument_type_dictionary=argument_dict)
        def cleaning(source: Path) -> Path:
            # A function that returns a new data frame after performing a cleaning step.
            ....
        ```
    """

    if func is None:
        return functools.partial(
            pipeline,
            pipeline_id=pipeline_id,
            external_run_id=external_run_id,
            external_run_uri=external_run_uri,
            editable=editable,
            argument_type_dictionary=argument_type_dictionary,
            return_var_name=return_var_name,
        )

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if func is None:
            return None

        # TODO: Check if function has multiple decorations.
        all_args = dict(zip(inspect.getfullargspec(func).args, args))
        all_args = {**all_args, **kwargs}
        all_annotations = func.__annotations__
        if argument_type_dictionary:
            all_annotations = {**all_annotations, **argument_type_dictionary}

        func_docstring_params = {}
        try:
            func_docstring = parse(text=func.__doc__ or "")
            func_docstring_params = {
                param.arg_name: param.description for param in func_docstring.params
            }
        except:
            logger.warning(
                f"Failed to parse docstring for func {func.__name__}. Artifact description will be missing"
            )

        radar_obj = RadarPipeline(
            pipeline_id=pipeline_id,
            external_run_uri=external_run_uri,
            external_run_id=external_run_id,
            editable=editable,
        )
        radar_obj.start_run()

        for arg_name, arg_value in all_args.items():
            if arg_name in all_annotations:
                typ = all_annotations[arg_name]
                description = (
                    func_docstring_params[arg_name]
                    if arg_name in func_docstring_params
                    else ""
                )
                _log_artifact(
                    radar_obj,
                    artifact_name=arg_name,
                    typ=typ,
                    value=arg_value,
                    description=description or "",
                )

        value = func(*args, **kwargs)

        # Check the return data path
        if "return" in all_annotations:
            typ = all_annotations["return"]
            description = (
                func_docstring.returns.description
                if func_docstring and func_docstring.returns
                else ""
            )
            _log_artifact(
                radar_obj,
                artifact_name=return_var_name or func.__name__,
                typ=typ,
                value=value,
                description=description or "",
                is_return_value=True,
            )

        radar_obj.finish_run()

        return value

    return wrapper


def _log_artifact(
    radar_obj: RadarPipeline,
    artifact_name: str,
    typ: Any,
    value: Any,
    description: str,
    is_return_value: bool = False,
) -> None:
    if not is_radar_type(typ):
        return

    artifact_type = None
    if is_dataset_artifact(typ):
        artifact_type = radar_types.Dataset
    elif is_model_artifact(typ):
        artifact_type = radar_types.Model
    elif is_artifact(typ):
        artifact_type = radar_types.Artifact

    if artifact_type is None:
        return

    radar_obj.log_artifact(
        uri=value,
        artifact_type=artifact_type,
        name=artifact_name,
        description=description,
    )



"""Radar Tool"""

from __future__ import annotations
from ._version import __version__

from pathlib import Path

import logging
import logging.config
from logging.handlers import RotatingFileHandler

from airadar import typing as radar_types
from airadar.tracking.decorators import pipeline
from airadar.extractors.dependency import Extractor
from airadar.extractors.dependency import PackageDependencyType
from airadar.extractors.dependency import DependencyExtractor
from airadar.radar import Radar, RadarPipeline
from airadar.tracking.fluent import log_model, log_deployment, active_run

from airadar.utils.config import airadar_configs, ConfigKeys

__all__ = [
    "radar_types",
    "pipeline",
    "Extractor",
    "Radar",
    "RadarPipeline",
    "PackageDependencyType",
    "DependencyExtractor",
    "log_model",
    "log_deployment",
    "active_run",
    "__version__",
]


def setup_logging() -> None:
    config_file = Path().resolve().joinpath("logging.conf")

    log_levels = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARN": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    }

    if config_file.exists() and config_file.is_file():
        logging.config.fileConfig(config_file, disable_existing_loggers=False)
    else:
        log_level_config = airadar_configs.get_value(ConfigKeys.AIRADAR_LOG_LEVEL)
        if not log_level_config:
            log_level_config = "INFO"
        else:
            log_level_config = str(log_level_config)

        log_level = log_levels.get(log_level_config, logging.INFO)

        rfh = RotatingFileHandler("radar.log", maxBytes=1 * 1024 * 1024, backupCount=3)
        rfh.setLevel(log_level)

        formatter = logging.Formatter(
            "%(asctime)s loglevel=%(levelname)-6s logger=%(name)s %(funcName)s() L%(lineno)-4d %(message)s"
        )
        rfh.setFormatter(formatter)

        logger = logging.getLogger("airadar")
        logger.setLevel(log_level)
        logger.addHandler(rfh)


setup_logging()



import os
import logging
import logging.config
from datetime import datetime
from contextlib import contextmanager
from airadar import typing as radar_types
from typing import Union, Optional, List, Generator, Dict

from airadar.typing import RadarResponse
from airadar.radar import RadarPipeline, Radar
from airadar.tracking.model import Dependency, ExternalUrlItem

disable_datacollection_env_var = "DISABLE_RADAR_COLLECTION"
external_run_id_env_var = "AIRADAR_EXTERNAL_RUN_ID"
logger = logging.getLogger(__name__)

_global_radar_obj: Optional[Radar] = None


def log_model(
    uri: str,
    name: Union[str, None] = None,
    description: Optional[str] = None,
    version: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
) -> RadarResponse:
    """
    Log a model with associated files without a pipeline. When creating artifacts
    within a pipeline use [active_run][airadar.tracking.fluent.active_run]

    Arguments:
        uri: URI to the model file, or the root directory, or glob pattern on local filesystem or cloud storage
            If a directory path is specified, all its contents will be recursively associated with this model.
        name: Human-readable name of the model this file is a part of.
            If None, the absolute filepath (for single file) or directory name (for directory path) is used.
        description: Human-readable description of the model. This is displayed on the Radar's model pages.
        version: Either a version string to be applied to all files associated with the logged model or a function which accepts a
            file's absolute path and returns a version string. If None, the SDK will attempt to infer the version for S3 resources.
        properties: Arbitrary key value properties that can be sent to the Radar console.

    Returns:
        A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
            Radar API call to log the model version
    """
    global _global_radar_obj
    if _global_radar_obj is None:
        _global_radar_obj = Radar()

    return _global_radar_obj.log_artifact(
        uri=uri,
        name=name,
        artifact_type=radar_types.Model,
        description=description,
        version=version,
        properties=properties,
    )


def log_deployment(
    application_id: str,
    application_version: str,
    environment: str,
    application_url: Optional[str] = None,
    build_url: Optional[str] = None,
    radar_uris: Optional[List[str]] = None,
    build_time: datetime = datetime.now(),
) -> RadarResponse:
    """
    Creates a new deployment against an application.

    Deployments are used to monitor deployed models within applications running in a user's organization.
    These applications might include, but aren't limited to, inference applications that
    serve the model to other users or software.

    Ideally, you should integrate this call into your CI/CD pipeline responsible for deploying the model.
    The CI/CD pipelines can provide many of the arguments to this function.
    This information can then allow Radar to trace back the provenance of the model within your application

    Note: application_id is Radar specific unique identifier obtained from onboarding the
    application on Radar web console.

    Arguments:
        application_id: The uuid of the application obtained from Radar's console.
        application_version: The version of the application within users' environment.
        environment: The name of the environment e.g. production, staging, dev etc.
        application_url: The URI to identify the application.
        build_url: This is typically the URI of the CI/CD pipeline that is deploying the call.
        radar_uris: The URI of the model in the form of radar://model/<model-name>/<model-version>.
        build_time: The date and time at which the application was built e.g. injected by your CI/CD pipeline.

    Returns:
        A [RadarResponse][airadar.typing.RadarResponse] object with optional metadata from the
            Radar API call to log the deployment

    Example:
        ```
        import airadar
        airadar.log_deployment(
            application_id="..",
            application_version="1.0.5",
            environment="production",
            application_url="https://example.com/api/v1/service",
            build_url="https://github.com/org/repo/actions/run/4",
            radar_uris=["radar://model/cost-predictions/latest"],
        )
        ```
    """
    global _global_radar_obj
    if _global_radar_obj is None:
        _global_radar_obj = Radar()

    return _global_radar_obj.log_deployment(
        application_id=application_id,
        application_version=application_version,
        environment=environment,
        application_url=application_url,
        build_url=build_url,
        radar_uris=radar_uris,
        build_time=build_time,
    )


@contextmanager
def active_run(
    pipeline_id: Optional[str] = os.environ.get("AIRADAR_PIPELINE_ID"),
    external_run_id: Optional[str] = None,
    external_run_uri: Optional[str] = None,
    editable: bool = False,
    dependencies: Optional[List[Dependency]] = None,
    pipeline_name: Optional[str] = None,
    external_url: Optional[ExternalUrlItem] = None,
) -> Generator[RadarPipeline, None, None]:
    """
    Initialize a pipeline session to log ML artifacts dataset(s) and model(s) generated within a pipeline.

    Models and datasets logged within the context of an active run are associated with the pipeline whose ID was passed in to the
    active_run function. If no pipeline ID is passed, the ```AIRADAR_PIPELINE_ID``` environment variable will be used instead.

    The external_run_id can be used to tie together distributed runs of a pipeline. It tells Radar
    to combine metadata generated across several processes into a single pipeline run. You can pass
    the external run id through the active_run method or by setting ```AIRADAR_EXTERNAL_RUN_ID``` environment variable.

    In general, values provided explicitly to the active_run method will override values provided via environment variables.

    Arguments:
        pipeline_id: The uuid of the pipeline obtained from Radar's web dashboard.
        external_run_id: ID for a pipeline run provided by a 3rd party tool (e.g. MLflow or Kubeflow).
        external_run_uri: URI identifying this pipeline run from the platform it ran on.
        editable: To keep pipeline open to log additional artifacts to it, set this to True.
        dependencies: List of dependencies that the pipeline run depends on.
                        If not provided, the SDK will attempt to read python environment for dependencies
        pipeline_name: Name of the pipeline. It is used to create a new pipeline if pipeline_id is not provided.
        external_url: Link to external source of the pipeline.

    Returns:
        A [RadarPipeline][airadar.radar.RadarPipeline] object that defines interface into calling

    Example:
        ```
        import airadar
        with airadar.active_run(pipeline_id="....") as radar:
            radar.log_model(..)
            radar.log_dataset(..)
        ```
    """
    radar = None
    try:
        radar = RadarPipeline(
            pipeline_id=pipeline_id,
            editable=editable,
            external_run_id=external_run_id,
            external_run_uri=external_run_uri,
            dependencies=dependencies,
            pipeline_name=pipeline_name,
            external_url=external_url,
        )
        radar.start_run()
        yield radar
    except Exception as e:
        logger.error(
            "Unexpected exception during Radar pipeline active run: %s", str(e)
        )
        if radar:
            radar.finish_run(failure=True)

        raise e
    finally:
        if radar and not radar.is_closed():
            radar.finish_run()



import abc
import logging
import threading
from typing import List, Union, Dict, Optional, Any
from urllib.parse import urlparse

from airadar.tracking.model import (
    File,
    VersionStrategy,
    VersionItem1,
    StorageTypeItem,
    ResourceNameItem,
)

logger = logging.getLogger(__name__)

# Each thread gets its own storage instance. This offers a simplest way
# to avoid issues with boto3 and thread safety. The only issue is a bucket
# blocked in one thread could be re-tried again in a different thread. However, that
# should not cause too much overhead.
schema_to_storage_mapping = threading.local()


class URIComponents:
    schema: Optional[str] = None
    bucket: Optional[str] = None
    path: Optional[str] = None


class CloudStorage(abc.ABC):
    def __init__(self) -> None:
        self._blocked_buckets: set[str] = set()

    @abc.abstractmethod
    def get_files_for_uri(self, uri: str) -> Union[List[str], None]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_file_metadata(
        self,
        uri: str,
    ) -> Union[File, None]:
        raise NotImplementedError

    @staticmethod
    def parse_uri(uri: str) -> URIComponents:
        components = URIComponents()
        parsed_uri = urlparse(uri)
        components.schema = parsed_uri.scheme

        components.bucket = parsed_uri.netloc
        path = parsed_uri.path
        path = path.lstrip("/")
        components.path = path

        return components

    @staticmethod
    def get_storage_obj_for_uri(uri: str) -> Union[None, "S3Storage", "GCStorage"]:
        components = CloudStorage.parse_uri(uri)
        if not components.schema:
            return None

        return CloudStorage.get_storage_obj_for_uri_schema(components.schema)

    @staticmethod
    def get_storage_obj_for_uri_schema(
        schema: str,
    ) -> Union[None, "S3Storage", "GCStorage"]:
        if schema == "s3":
            if not hasattr(schema_to_storage_mapping, "s3_storage_obj"):
                s3_storage_obj = S3Storage()
                schema_to_storage_mapping.s3_storage_obj = s3_storage_obj

            if isinstance(schema_to_storage_mapping.s3_storage_obj, S3Storage):
                return schema_to_storage_mapping.s3_storage_obj
        elif schema == "gs":
            if not hasattr(schema_to_storage_mapping, "gc_storage_obj"):
                gs_storage_obj = GCStorage()
                schema_to_storage_mapping.gs_storage_obj = gs_storage_obj

            if isinstance(schema_to_storage_mapping.gs_storage_obj, GCStorage):
                return schema_to_storage_mapping.gs_storage_obj

        return None


class S3Storage(CloudStorage):
    def __init__(self) -> None:
        super().__init__()
        self.s3_client = None
        self.s3fs_client = None
        self._botocore = None

        self.s3_glob_available = True
        try:
            import boto3
            import botocore

            session = boto3.Session()
            self.s3_client = session.client("s3")
            self._botocore = botocore
        except ImportError as e:
            logger.error(
                """
                Failed to import boto3 module. Radar will not be able to fetch extra metadata for S3 buckets.\n
                You can install boto3 support using either: \n
                1. pip install boto3 \n
                2. pip install radar[aws]\n
                %s
                """,
                str(e),
            )

        try:
            import s3fs

            self.s3fs_client = s3fs.S3FileSystem()
        except:
            logger.info(
                "Failed to import s3fs module. Glob patterns for S3 will not be supported."
            )

    @staticmethod
    def file_from_s3_metadata(uri: str, response: Dict[str, Any]) -> File:
        f = File(uri=uri)

        f.size = response.get("ContentLength")
        version_id = response.get("VersionId")
        f.version = VersionItem1(root=str(version_id)) if version_id else None
        f.storage_type = StorageTypeItem(root="S3")
        f.version_strategy = VersionStrategy.s3

        return f

    def _get_files_for_glob(self, bucket: str, path: str) -> Union[None, List[str]]:
        if self.s3fs_client is None:
            return None

        try:
            detail_files = self.s3fs_client.glob(f"{bucket}/{path}", detail=True)
        except Exception as e:
            # s3fs swallows all exceptions from S3 (e.g. AccessDenied), so this only capture errors
            # from NoCredentialsError and other unexpected errors
            logger.error(
                "Unexpected error while trying to list files for glob pattern %s/%s: %s",
                bucket,
                path,
                str(e),
            )
            return None

        # This returns the fully qualified S3 path while excluding directories returned by glob
        return [
            f"s3://{key}"
            for key, details in detail_files.items()
            if details["type"] == "file"
        ]

    def _get_s3_uri(self, bucket: str, path: str) -> str:
        return f"s3://{bucket}/{path}"

    def get_files_for_uri(self, uri: str) -> Union[List[str], None]:
        if not self.s3_client:
            return None

        components = CloudStorage.parse_uri(uri)
        schema = components.schema
        bucket = components.bucket
        path = components.path

        logger.debug(
            "Processing request for S3: schema=%s, bucket=%s, path=%s",
            schema,
            bucket,
            path,
        )

        if schema != "s3":
            logger.debug(f"Invalid URI schema for an S3 path: %s")
            return None

        if not bucket:
            logger.debug(f"Invalid URI schema for an S3 path: %s")
            return None

        if bucket in self._blocked_buckets:
            logger.debug(
                f"Requested bucket %s is blacklisted because of previous error in accessing it",
                components.bucket,
            )
            return None

        path = components.path or ""

        # If a customer wants to add files with a literal "*" or "?" in their keys,
        # this approach may cause unexpected behavior.
        # Discussed in https://github.com/protectai/airadar/issues/78
        if "*" in path or "?" in path or ("[" in path and "]" in path):
            return self._get_files_for_glob(bucket, path)

        files: List[str] = []
        continuation_token = None

        try:
            while True:
                if continuation_token is None:
                    response = self.s3_client.list_objects_v2(
                        Bucket=bucket,
                        Prefix=path,
                    )
                    if not "Contents" in response:
                        logger.warning("No files found in S3 for URI %s", uri)
                        return files
                else:
                    response = self.s3_client.list_objects_v2(
                        Bucket=bucket,
                        Prefix=path,
                        ContinuationToken=continuation_token,
                    )

                content_list = response.get("Contents", [])
                files.extend(
                    [
                        self._get_s3_uri(bucket, content["Key"])
                        for content in content_list
                    ]
                )

                # NextContinuationToken is only defined in response if IsTruncated is True
                # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3/client/list_objects_v2.html
                if response.get("IsTruncated", False) == False:
                    break
                else:
                    continuation_token = response.get("NextContinuationToken", None)

            return files
        except Exception as e:
            if (
                self._botocore
                and isinstance(e, self._botocore.exceptions.ClientError)
                and e.response["Error"]["Code"] == "AccessDenied"
            ):
                self._blocked_buckets.add(bucket)
                logger.error(
                    "Access denied for URI %s. Adding bucket %s to block list.",
                    uri,
                    bucket,
                )
                return None

            logger.error(
                "Unexpected error while trying to list files for URI %s: %s",
                uri,
                str(e),
            )
            return None

    def get_file_metadata(self, uri: str) -> Union[File, None]:
        if not self.s3_client:
            return None

        components = self.parse_uri(uri)

        bucket = components.bucket
        path = components.path

        if not bucket or not path:
            return None

        try:
            response = self.s3_client.head_object(Bucket=bucket, Key=path)
            metadata = S3Storage.file_from_s3_metadata(uri, response)
            metadata.resource_name = ResourceNameItem(root=bucket)
            return metadata

        except Exception as e:
            if (
                self._botocore
                and isinstance(e, self._botocore.exceptions.ClientError)
                and e.response["Error"]["Code"] == "AccessDenied"
            ):
                self._blocked_buckets.add(bucket)
                logger.error(
                    "Access denied for URI %s. Adding bucket %s to block list.",
                    uri,
                    bucket,
                )
                return None

            logger.error(
                "Unexpected error while trying to retrieve file metadata for URI %s: %s",
                uri,
                str(e),
            )
            return None


class GCStorage(CloudStorage):
    def __init__(self) -> None:
        super().__init__()
        self.storage_client = None
        _gcs_sdk_available = False
        try:
            from google.cloud import storage

            _gcs_sdk_available = True
        except ImportError as e:
            logger.error(
                """
                Failed to import google-cloud-storage module. Radar will not be able to fetch extra metadata for GS buckets.\n
                You can install google-cloud-storage support using either: \n
                1. pip install google-cloud-storage \n
                2. pip install radar[gc]\n
                %s
                """,
                str(e),
            )

        if _gcs_sdk_available:
            from google.auth.exceptions import DefaultCredentialsError

            try:
                self.storage_client = storage.Client()
            except DefaultCredentialsError as e:
                logger.error(
                    """
                    Failed to authenticate with google cloud. Metadata from GCS will not be available.\n
                    You can set the path to service account json file by setting the environment variable for GOOGLE_APPLICATION_CREDENTIALS_PATH.
                    %s
                    """,
                    str(e),
                )
            except Exception as e:
                logger.error(
                    "Unknown error when trying to create google storage client: %s",
                    str(e),
                )

    @staticmethod
    def file_from_gcs_metadata(uri: str, response: Any) -> File:
        f = File(uri=uri)
        f.resource_name = ResourceNameItem(root=response.bucket.name)
        f.size = response.size
        f.storage_type = StorageTypeItem(root="GCS")
        if response.md5_hash:
            f.version = VersionItem1(root=response.md5_hash)
            f.version_strategy = (
                VersionStrategy.custom
            )  # TODO Add md5 on the API side first
        elif response.crc32c:
            f.version = VersionItem1(root=response.crc32c)
            f.version_strategy = (
                VersionStrategy.custom
            )  # TODO Add crc32 on the API side first
        else:
            f.version = VersionItem1(root=response.etag)
            f.version_strategy = (
                VersionStrategy.custom
            )  # TODO Add gsc on the API side first

        return f

    def get_file_metadata(self, uri: str) -> Union[File, None]:
        if not self.storage_client:
            return None
        from google.cloud.exceptions import NotFound, Forbidden
        from google.cloud.storage import Blob

        components = self.parse_uri(uri)

        bucket = components.bucket
        path = components.path

        if not bucket or not path:
            return None

        try:
            bucket_obj = self.storage_client.get_bucket(bucket)
            blob = bucket_obj.get_blob(path)
            if isinstance(blob, Blob):
                return GCStorage.file_from_gcs_metadata(uri, blob)

        except NotFound as e:
            logger.error(
                "Bucket not found for uri: %s\n%s",
                uri,
                str(e),
            )
            self._blocked_buckets.add(bucket)
            return None
        except Forbidden as e:
            logger.error("403 Forbidden for uri: %s\n%s", uri, str(e))
            self._blocked_buckets.add(bucket)
            return None
        except Exception as e:
            logger.error("Unknown error for uri: %s\n%s", uri, str(e))
            self._blocked_buckets.add(bucket)
            return None

        return None

    def get_files_for_uri(self, uri: str) -> Union[List[str], None]:
        if not self.storage_client:
            return None

        from google.cloud.exceptions import NotFound, Forbidden

        components = CloudStorage.parse_uri(uri)
        schema = components.schema
        bucket = components.bucket
        path = components.path

        logger.debug(
            "Processing request for GCS: schema=%s, bucket=%s, path=%s",
            schema,
            bucket,
            path,
        )

        if schema != "gs":
            logger.debug(f"Invalid URI schema for an GS path: %s")
            return None

        if not bucket:
            logger.debug(f"Invalid URI schema for an GS path: %s")
            return None

        if bucket in self._blocked_buckets:
            logger.debug(
                f"Requested bucket %s is blacklisted because of previous error in accessing it",
                components.bucket,
            )
            return None

        path = components.path or ""

        files: List[str] = []

        try:
            bucket_obj = self.storage_client.get_bucket(bucket)
            # If a customer wants to add files with a literal "*" or "?" in their keys,
            # this approach may cause unexpected behavior.
            # Discussed in https://github.com/protectai/airadar/issues/78
            ret = None
            if "*" in path or "?" in path or ("[" in path and "]" in path):
                ret = bucket_obj.list_blobs(match_glob=path)
            else:
                ret = bucket_obj.list_blobs(prefix=path)

            if not ret:
                logger.warning("No files found in GS for URI %s", uri)
                return files

            # GCS SDK returns an iterator that seems to handle pagination.
            for blob in ret:
                files.append(f"gs://{blob.bucket.name}/{blob.name}")

            return files
        except Forbidden as e:
            logger.error("403 Forbidden for uri: %s\n%s", uri, str(e))
            self._blocked_buckets.add(bucket)
            return None
        except NotFound as e:
            logger.error(
                "Requested bucket for %s was not found\n%s",
                uri,
                str(e),
            )
            self._blocked_buckets.add(bucket)
            return None

        except Exception as e:
            logger.error(
                "Unexpected error while trying to list files for URI %s\n%s",
                uri,
                str(e),
            )
            return None

        return None



import logging
import threading
import time
import json
import os
from datetime import datetime

from typing import Optional, List

import airadar
from airadar.radar import Radar

logger = logging.getLogger(__name__)

mlflow_tracking_available = True

try:
    import mlflow
    from mlflow.tracking import MlflowClient
    from mlflow.entities import Run
except ImportError:
    logger.error(
        "Failed to import mlflow, tracking of mlflow artifacts will be disabled"
    )
    mlflow_tracking_available = False


def _find_all_model_artifacts(artifact_uri: str, artifacts: List[str]) -> List[str]:
    ml_model_paths = [path for path in artifacts if "MLmodel" in path]
    ret = []
    for ml_model_path in ml_model_paths:
        if "/" not in ml_model_path:
            continue
        ret.append(f"{artifact_uri}/{ml_model_path.split('/')[0]}")

    return ret


def _track_mlflow_current_run(run_id: str, timestamp: datetime) -> None:
    if run_id is None:
        return

    logger.info(f"AIRadar is now tracking the run: {run_id}")

    mlflow_run = None
    while True:
        mlflow_run = MlflowClient().get_run(run_id)
        if not mlflow_run:
            logger.warn(
                "Mlflow run_id was not valid. No further tracking will be done."
            )
            return

        if mlflow_run.info.status != "RUNNING":
            break

        time.sleep(5)  # TODO make this time configurable

    radar_pipeline_id = os.environ.get("AIRADAR_PIPELINE_ID")
    if not radar_pipeline_id:
        logger.warn(
            "A Radar pipeline ID is missing. Models from Mlflow will be logged without a pipeline"
        )
        _log_model_helper(Radar(), mlflow_run)
    else:
        with airadar.active_run(
            pipeline_id=radar_pipeline_id, external_run_id=mlflow_run.info.run_id
        ) as radar:
            _log_model_helper(radar, mlflow_run)
            for dataset_input in mlflow_run.inputs.dataset_inputs:
                source = dataset_input.dataset.source
                dataset_info = None
                try:
                    if source:
                        dataset_info = json.loads(source)
                except:
                    logger.info(
                        f"Got MLFlow dataset with source that we couldn't understand {source}"
                    )

                dataset_uri = None
                if dataset_info:
                    dataset_uri = dataset_info.get("url") or dataset_info.get("uri")

                if dataset_uri:
                    logger.info(
                        f"Logging a dataset from mlflow with uri = {dataset_uri}"
                    )
                    radar.log_dataset(dataset_uri, name=dataset_input.dataset.name)


def _log_model_helper(radar: Radar, mlflow_run: Run) -> None:
    model_versions = MlflowClient().search_model_versions(
        filter_string=f"run_id='{mlflow_run.info.run_id}'"
    )
    for model_version in model_versions:
        model_uri = f"models:/{model_version.name}/{model_version.version}"
        logger.info(
            f"Logging mlflow model from run {mlflow_run.info.run_id} with uri {model_uri}"
        )
        radar.log_model(
            model_uri, name=model_version.name, version=str(model_version.version)
        )


def _get_mlflow_run_id() -> None:
    run_id = None
    start_time = datetime.now()
    while True:
        if not mlflow.active_run() and not mlflow.last_active_run():
            time.sleep(1)  # TODO make this time configurable
        else:
            last_active_run = mlflow.last_active_run()
            if last_active_run:
                run_id = last_active_run.info.run_id
            break

    if run_id:
        logger.info("Got the mlflow run_id from the environment")
        tracking_thread = threading.Thread(
            target=_track_mlflow_current_run, args=(run_id, start_time), daemon=False
        )
        tracking_thread.start()


def start_mlflow_tracking(run_id: Optional[str] = None) -> None:
    """
    Track artifacts being logged to mlflow.

    if run_id is provided, the SDK will start tracking artifacts for the given run_id
    If run_id is not provided, the SDK waits until run_id of current run is available and then launch
        tracking of artifacts.

    Arguments:
        run_id: MLflow run id

    """

    # We start tracking with two threads, one is daemon thread that runs if there is no
    # run id provided, this thread then launches our second thread to gather artifacts.
    # If run_id is provided, we launch only the tracking thread.
    if not mlflow_tracking_available:
        logger.warn("Failed to import mlflow. Tracking is disabled")
        return

    if not run_id:
        # Temporarily making this a non-daemon thread. For very short mlflow code,
        # a daemon thread won't even get a chance to get the run-id. Since start_mlflow_tracking()
        # is only being called through mlflow run_context we will get the mlflow run_id eventually
        # TODO: Later can make this little more robust to do a final context flush if
        #       start_mlflow_tracking() was called.
        run_id_thread = threading.Thread(target=_get_mlflow_run_id, daemon=False)
        run_id_thread.start()
    else:
        tracking_thread = threading.Thread(
            target=_track_mlflow_current_run, args=(run_id,), daemon=False
        )
        tracking_thread.start()



import logging
from typing import Optional, List, Generator, Union
from urllib.parse import urlparse

mlflow_tracking_available = True

logger = logging.getLogger(__name__)

try:
    from mlflow.tracking import MlflowClient
    from mlflow.entities.model_registry import ModelVersion
except ImportError:
    logger.error(
        "Failed to import mlflow, tracking of mlflow artifacts will be disabled"
    )
    mlflow_tracking_available = False

from airadar.tools.cloud_storage import CloudStorage


def _find_all_model_artifacts(artifact_uri: str, artifacts: List[str]) -> List[str]:
    ml_model_paths = [path for path in artifacts if "MLmodel" in path]
    ret = []
    for ml_model_path in ml_model_paths:
        if "/" not in ml_model_path:
            continue
        ret.append(f"{artifact_uri}/{ml_model_path.split('/')[0]}")

    return ret


def _yield_mlflow_artifacts(
    client: MlflowClient, run_id: str, path: Optional[str] = None
) -> Generator[str, None, None]:
    for item in client.list_artifacts(run_id, path):
        if item.is_dir:
            yield from _yield_mlflow_artifacts(client, run_id, item.path)
        else:
            yield item.path


def get_mlflow_model_files(run_id: str) -> Optional[List[str]]:
    if not mlflow_tracking_available:
        logger.warn("Mlflow tracking is not available. Please install mlflow package.")
        return None

    if not run_id:
        logger.error("Need a valid run_id for fetching model(s) files from Mlflow")
        return None

    try:
        client = MlflowClient()
        mlflow_run = client.get_run(run_id)
        if not mlflow_run:
            logger.info(f"Mlflow run was not found {run_id}")
            return None

        artifacts = list(_yield_mlflow_artifacts(client, run_id))
        model_artifacts_uris = _find_all_model_artifacts(
            mlflow_run.info.artifact_uri, artifacts
        )

        files: List[str] = []
        for artifact_uri in model_artifacts_uris:
            storage_obj = CloudStorage.get_storage_obj_for_uri(artifact_uri)
            if storage_obj:
                file_list = storage_obj.get_files_for_uri(artifact_uri)
                if file_list:
                    files.extend(file_list)

        return files
    except Exception as e:
        logger.error(
            "Failed to fetch files for run_id %s: %s", run_id, e, exc_info=True
        )

    return None


def get_mlflow_model_info(mlflow_uri: str) -> Union["ModelVersion", None]:
    if not mlflow_tracking_available:
        logger.warning(
            "Mlflow tracking is not available. Please install mlflow package."
        )
        return None

    parsed_uri = urlparse(mlflow_uri)
    if parsed_uri.scheme not in ["runs", "models"]:
        return None

    mlflow_model_version = None

    mlflow_client = MlflowClient()

    if parsed_uri.scheme == "runs":
        run_id = None
        parts = parsed_uri.path.split("/")
        if len(parts) > 2:
            run_id = parts[1]

        if not run_id:
            logger.error(
                f"Mlflow URI {mlflow_uri} was not valid. It should be of format runs:/<run-id>/<run-name>"
            )
            return None

        model_versions = mlflow_client.search_model_versions(
            filter_string=f"run_id='{run_id}'"
        )
        # TODO: With runs we can have multiple models. The upstream code
        # need some modifications to handle this case. For now, we just picking
        # one model. Our Mlflow tracking plugin is handling multiple models in a run.
        if model_versions:
            mlflow_model_version = model_versions[0]
    else:
        # Only other possibility now is 'models:/' URI.
        name, version = None, None
        if "@" in parsed_uri.path:
            parts = parsed_uri.path.split("@")
            name, alias = parts[0][1:], parts[1]
            if name and alias:
                mlflow_model_version = mlflow_client.get_model_version_by_alias(
                    name=name, alias=alias
                )
            else:
                logger.debug("Failed to parse model name and alias from Mlflow URI")
        else:
            parts = parsed_uri.path.split("/")
            name, version = None, None
            if len(parts) > 2:
                name, version = parts[1], parts[2]

            if name and version:
                mlflow_model_version = mlflow_client.get_model_version(
                    name=name, version=version
                )
            else:
                logger.debug("Failed to parse model name and version from Mlflow URI")

    return mlflow_model_version



import airadar.typing as radar_types

from typing import Any


def has_annotation_metadata(typ: Any) -> bool:
    if not hasattr(typ, "__metadata__"):
        return False

    return True


def is_input_artifact(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    return bool(typ.__metadata__[0] == radar_types.InputAnnotation)


def is_output_artifact(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    return bool(typ.__metadata__[0] == radar_types.OutputAnnotation)


def is_dataset_artifact(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    return bool(typ.__metadata__[0] == radar_types.DatasetType)


def is_model_artifact(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    return bool(typ.__metadata__[0] == radar_types.ModelType)


def is_artifact(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    return bool(typ.__metadata__[0] == radar_types.ArtifactType)


def is_radar_type(typ: Any) -> bool:
    if not has_annotation_metadata(typ):
        return False

    all_radar_types = [
        radar_types.InputAnnotation,
        radar_types.OutputAnnotation,
        radar_types.DatasetType,
        radar_types.ModelType,
        radar_types.ArtifactType,
    ]

    return bool(typ.__metadata__[0] in all_radar_types)



import abc
import pathlib
import json
import logging
import importlib.metadata

from enum import Enum
from typing import Optional, Any, Dict

logger = logging.getLogger(__name__)


class Extractor(abc.ABC):
    @abc.abstractmethod
    def get_report(self) -> object:
        raise NotImplementedError


class PackageDependencyType(Enum):
    ENVIRONMENT = 1
    PIP_LOCK_FILE = 2
    POETRY_LOCK_FILE = 3
    REQUIREMENTS_TXT_FILE = 4
    CONDA_LIST_FILE = 5
    CONDA_EXPLICIT_LIST_FILE = 6


class SimpleEnvironmentParser:
    """
    A simplified parser that extracts package dependency information from
    the local Python environment.
    """

    def parse(self) -> Dict[str, Any]:
        """
        Parse the current Python environment and return information in a format
        compatible with the expected output structure.

        Returns:
            Dict containing package information in the expected format.
        """
        components = []

        # Use pkg_resources to get installed packages
        for dist in importlib.metadata.distributions():
            try:
                # Get package name and version
                package_name = dist.metadata["Name"]
                package_version = dist.version

                component: Dict[str, Any] = {
                    "type": "library",
                    "name": package_name,
                    "version": package_version,
                    "purl": f"pkg:pypi/{package_name}@{package_version}",
                    "author": "",
                    "licenses": [],
                }

                if "Author" in dist.metadata:
                    component["author"] = dist.metadata["Author"]
                elif "Author-email" in dist.metadata:
                    # Fall back to author email if author name is not available
                    component["author"] = (
                        dist.metadata["Author-email"].split("<")[0].strip()
                    )

                # Extract license information
                license_info = None
                if "License" in dist.metadata:
                    license_info = dist.metadata["License"]
                elif "Classifier" in dist.metadata:
                    # Try to extract license from classifiers if direct license not available
                    for classifier in dist.metadata.get_all("Classifier", []):
                        if classifier.startswith("License ::"):
                            license_parts = classifier.split(" :: ")
                            if len(license_parts) > 2:
                                license_info = license_parts[-1]
                                break

                # Format license information in the expected structure
                if license_info:
                    component["licenses"] = [{"name": license_info}]

                components.append(component)
            except Exception as e:
                logger.debug(f"Failed to extract package information for {dist}: {e}")
                continue

        return {"components": components}


class DependencyExtractor(Extractor):
    def __init__(
        self,
        package_type: PackageDependencyType,
        dependency_file_path: Optional[pathlib.Path] = None,
    ):
        if package_type != PackageDependencyType.ENVIRONMENT:
            raise NotImplementedError("Only environment extractor is available.")

        self._package_type = package_type
        self._dependency_file_path = dependency_file_path
        self.components = None

    def get_report(self) -> Dict[str, Any]:
        """
        Get dependency report.

        Returns:
            Dict containing the package information.
        """
        if self.components:
            return self.components

        parser = SimpleEnvironmentParser()
        dependency_report = parser.parse()

        return dependency_report
